"""
This module provides definitions and functionality related to encoding monetary values and operating on them.
"""

__all__ = [
    "IncompatibleCurrencyError",
    "MonetaryOperationException",
    "Money",
    "NoMoney",
    "NoPrice",
    "NoneMoney",
    "NonePrice",
    "Price",
    "SomeMoney",
    "SomePrice",
]

from abc import abstractmethod
from decimal import Decimal, DivisionByZero, InvalidOperation
from typing import Any, NamedTuple, Optional, Union, overload

from .commons.errors import ProgrammingError
from .commons.numbers import Numeric
from .commons.zeitgeist import Date
from .currencies import Currency
from .exchange import FXRateLookupError, FXRateService


class IncompatibleCurrencyError(ValueError):
    """
    Provides an exception indicating that there is an attempt for performing monetary operations
    with incompatible currencies.
    """

    def __init__(self, ccy1: Currency, ccy2: Currency, operation: str = "<Unspecified>") -> None:
        """
        Initializes an incompatible currency error message.
        """
        ## Keep sloys:
        self.ccy1 = ccy1
        self.ccy2 = ccy2
        self.operation = operation

        ## Call super:
        super().__init__(f"{ccy1.code} vs {ccy2.code} are incompatible for operation '{operation}'.")


class MonetaryOperationException(TypeError):
    """
    Provides an exception that a certain monetary operation can not be carried on.
    """

    pass


class Money:
    """
    Provides an abstract money model and its semantics.
    """

    ## No need for slots.
    __slots__ = ()

    #: Defines the *undefined* money object as a singleton.
    NA: "Money"

    #: Returns the currency of the money object, if defined.
    #:
    #: Note that a :class:`TypeError` is raised if call-site attempts to access this property of an undefined money.
    ccy: Currency

    #: Returns the quantity of the money object, if defined.
    #:
    #: Note that a :class:`TypeError` is raised if call-site attempts to access this property of an undefined money.
    qty: Decimal

    #: Returns the value date of the money object, if defined.
    #:
    #: Note that a :class:`TypeError` is raised if call-site attempts to access this property of an undefined money.
    dov: Date

    #: Indicates that the money is a *defined* monetary value.
    defined: bool  # noqa: E704

    #: Indicates that the money is an *undefined* monetary value.
    undefined: bool

    @abstractmethod
    def is_equal(self, other: Any) -> bool:
        """
        Checks the equality of two money objects.

        In particular:

        1. ``True`` if ``other`` is a money object **and** all slots are same.
        2. ``False`` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def as_boolean(self) -> bool:
        """
        Returns the logical representation of the money object.

        In particular:

        1. ``False`` if money is *undefined* **or** money quantity is ``zero``.
        2. ``True`` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def as_float(self) -> float:
        """
        Returns the quantity as a ``float`` if *defined*, raises class:`MonetaryOperationException` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def as_integer(self) -> int:
        """
        Returns the quantity as an ``int`` if *defined*, raises class:`MonetaryOperationException` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def abs(self) -> "Money":
        """
        Returns the absolute money if *defined*, itself otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def negative(self) -> "Money":
        """
        Negates the quantity of the monetary value if *defined*, itself otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def positive(self) -> "Money":
        """
        Returns same monetary value if *defined*, itself otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def round(self, ndigits: int = 0) -> "Money":
        """
        Rounds the quantity of the monetary value to ``ndigits`` by using ``HALF_EVEN`` method if *defined*, itself
        otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def add(self, other: "Money") -> "Money":
        """
        Performs monetary addition on the money object and the given ``other`` money object.

        Note that::

        1. Raises :class:`IncompatibleCurrencyError` if currencies do not match.
        2. If any of the operands are undefined, returns the other one conveniently.
        3. Dates are carried forward as a result of addition of two defined money objects.
        """
        raise NotImplementedError

    @abstractmethod
    def scalar_add(self, other: Numeric) -> "Money":
        """
        Performs scalar addition on the quantity of the money.

        Note that undefined money object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def subtract(self, other: "Money") -> "Money":
        """
        Performs monetary subtraction on the money object and the given ``other`` money object.

        Note that::

        1. Raises :class:`IncompatibleCurrencyError` if currencies do not match.
        2. If any of the operands are undefined, returns the other one conveniently.
        3. Dates are carried forward as a result of addition of two defined money objects.
        """
        raise NotImplementedError

    @abstractmethod
    def scalar_subtract(self, other: Numeric) -> "Money":
        """
        Performs scalar subtraction on the quantity of the money.

        Note that undefined money object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def multiply(self, other: Numeric) -> "Money":
        """
        Performs scalar multiplication.

        Note that undefined money object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def divide(self, other: Numeric) -> "Money":
        """
        Performs ordinary division on the money object if *defined*, itself otherwise.

        Note that division by zero yields an undefined money object.
        """
        raise NotImplementedError

    @abstractmethod
    def floor_divide(self, other: Numeric) -> "Money":
        """
        Performs floor division on the money object if *defined*, itself otherwise.

        Note that division by zero yields an undefined money object.

        """
        raise NotImplementedError

    @abstractmethod
    def lt(self, other: "Money") -> bool:
        """
        Applies "less than" comparison against ``other`` money.

        Note that::

        1. Undefined money objects are always less than ``other`` if ``other`` is not undefined, and
        2. :class:`IncompatibleCurrencyError` is raised when comparing two defined money objects with different
        currencies.
        """
        pass

    @abstractmethod
    def lte(self, other: "Money") -> bool:
        """
        Applies "less than or equal to" comparison against ``other`` money.

        Note that::

        1. Undefined money objects are always less than or equal to ``other``, and
        2. :class:`IncompatibleCurrencyError` is raised when comparing two defined money objects with different
        currencies.
        """
        pass

    @abstractmethod
    def gt(self, other: "Money") -> bool:
        """
        Applies "greater than" comparison against ``other`` money.

        Note that::

        1. Undefined money objects are never greater than ``other``,
        2. Defined money objects are always greater than ``other`` if other is undefined, and
        3. :class:`IncompatibleCurrencyError` is raised when comparing two defined money objects with different
        currencies.
        """
        pass

    @abstractmethod
    def gte(self, other: "Money") -> bool:
        """
        Applies "greater than or equal to" comparison against ``other`` money.

        Note that::

        1. Undefined money objects are never greater than or equal to ``other`` if ``other`` is defined,
        2. Undefined money objects are greater than or equal to ``other`` if ``other is undefined, and
        3. :class:`IncompatibleCurrencyError` is raised when comparing two defined money objects with different
        currencies.
        """
        pass

    @abstractmethod
    def with_ccy(self, ccy: Currency) -> "Money":
        """
        Creates a new money object with the given currency if money is *defined*, returns itself otherwise.
        """
        pass

    @abstractmethod
    def with_qty(self, qty: Decimal) -> "Money":
        """
        Creates a new money object with the given quantity if money is *defined*, returns itself otherwise.
        """
        pass

    @abstractmethod
    def with_dov(self, dov: Date) -> "Money":
        """
        Creates a new money object with the given value date if money is *defined*, returns itself otherwise.
        """
        pass

    @abstractmethod
    def ccy_or(self, default: Currency) -> Currency:
        """
        Returns the ``ccy`` if the monetary value is *defined*, ``default`` otherwise.

        >>> from pypara.currencies import Currencies
        >>> somemoney = Money.of(Currencies["USD"], Decimal('1'), Date(2019, 1, 1))
        >>> somemoney.ccy_or(Currencies["EUR"]).code
        'USD'
        >>> nonemoney = Money.of(Currencies["USD"], None, None)
        >>> nonemoney.ccy_or(Currencies["EUR"]).code
        'EUR'
        """
        pass

    @abstractmethod
    def qty_or(self, default: Decimal) -> Decimal:
        """
        Returns the ``qty`` if the monetary value is *defined*, ``default`` otherwise.

        >>> from pypara.currencies import Currencies
        >>> somemoney = Money.of(Currencies["USD"], Decimal('1'), Date(2019, 1, 1))
        >>> somemoney.qty_or(Decimal(0))
        Decimal('1.00')
        >>> nonemoney = Money.of(None, Decimal('1'), None)
        >>> nonemoney.qty_or(Decimal(0))
        Decimal('0')
        """
        pass

    @abstractmethod
    def dov_or(self, default: Date) -> Date:
        """
        Returns the ``dov`` if the monetary value is *defined*, ``default`` otherwise.

        >>> from pypara.currencies import Currencies
        >>> somemoney = Money.of(Currencies["USD"], Decimal('1'), Date(2019, 1, 1))
        >>> somemoney.dov_or(Date(2001, 1, 1))
        datetime.date(2019, 1, 1)
        >>> nonemoney = Money.of(None, None, Date(2019, 1, 1))
        >>> nonemoney.dov_or(Date(2001, 1, 1))
        datetime.date(2001, 1, 1)
        """
        pass

    @abstractmethod
    def convert(self, to: Currency, asof: Optional[Date] = None, strict: bool = False) -> "Money":
        """
        Converts the monetary value from one currency to another.

        Raises :class:`FXRateLookupError` if no foreign exchange rate can be found for conversion.

        Note that we will carry the date forward as per ``asof`` date.
        """
        raise NotImplementedError

    @classmethod
    def of(cls, ccy: Optional[Currency], qty: Optional[Decimal], dov: Optional[Date]) -> "Money":
        """
        Provides a factory method to create a new money object in a safe manner.
        """
        if qty is None or ccy is None or dov is None:
            return NoMoney
        return SomeMoney(ccy, ccy.quantize(qty), dov)

    @property
    @abstractmethod
    def price(self) -> "Price":
        """
        Returns the price representation of the money object.
        """
        raise NotImplementedError

    @abstractmethod
    def __bool__(self) -> bool:
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __abs__(self) -> "Money":
        pass

    @abstractmethod
    def __float__(self) -> float:
        pass

    @abstractmethod
    def __int__(self) -> int:
        pass

    @overload
    def __round__(self) -> int:
        ...

    @overload
    def __round__(self, ndigits: None) -> int:
        ...

    @overload
    def __round__(self, ndigits: int) -> "Money":
        ...

    def __round__(self, ndigits: Optional[int] = 0) -> Union["Money", int]:
        return self.round(ndigits or 0)

    @abstractmethod
    def __neg__(self) -> "Money":
        pass

    @abstractmethod
    def __pos__(self) -> "Money":
        pass

    @abstractmethod
    def __add__(self, other: "Money") -> "Money":
        pass

    @abstractmethod
    def __sub__(self, other: "Money") -> "Money":
        pass

    @abstractmethod
    def __mul__(self, other: Numeric) -> "Money":
        pass

    @abstractmethod
    def __truediv__(self, other: Numeric) -> "Money":
        pass

    @abstractmethod
    def __floordiv__(self, other: Numeric) -> "Money":
        pass

    @abstractmethod
    def __lt__(self, other: "Money") -> bool:
        pass

    @abstractmethod
    def __le__(self, other: "Money") -> bool:
        pass

    @abstractmethod
    def __gt__(self, other: "Money") -> bool:
        pass

    @abstractmethod
    def __ge__(self, other: "Money") -> bool:
        pass


class SomeMoney(Money, NamedTuple("SomeMoney", [("ccy", Currency), ("qty", Decimal), ("dov", Date)])):
    """
    Provides a *defined* money object model.
    """

    __slots__ = ()

    defined = True

    undefined = False

    def is_equal(self, other: Any) -> bool:
        return other.__class__ is SomeMoney and tuple(self) == tuple(other)

    def as_boolean(self) -> bool:
        return self[1].__bool__()

    def as_float(self) -> float:
        return self[1].__float__()

    def as_integer(self) -> int:
        return self[1].__int__()

    def abs(self) -> "Money":
        c, q, d = self
        return SomeMoney(c, q.__abs__(), d)

    def negative(self) -> "Money":
        c, q, d = self
        return SomeMoney(c, q.__neg__(), d)

    def positive(self) -> "Money":
        c, q, d = self
        return SomeMoney(c, q.__pos__(), d)

    def round(self, ndigits: int = 0) -> "Money":
        c, q, d = self
        dec = c.decimals
        return SomeMoney(c, q.__round__(ndigits if ndigits < dec else dec), d)

    def add(self, other: "Money") -> "Money":
        if other.undefined:
            return self

        c1: Currency
        q1: Decimal
        d1: Date
        c2: Currency
        q2: Decimal
        d2: Date
        c1, q1, d1 = self
        c2, q2, d2 = other  # type: ignore

        if c1 != c2:
            raise IncompatibleCurrencyError(ccy1=c1, ccy2=c2, operation="addition")

        return SomeMoney(c1, q1 + q2, d1 if d1 > d2 else d2)

    def scalar_add(self, other: Numeric) -> "Money":
        ## TODO: **try** not casting other to Decimal.
        c, q, d = self
        return SomeMoney(c, (q + Decimal(other)).quantize(c.quantizer), d)

    def subtract(self, other: "Money") -> "Money":
        if other.undefined:
            return self

        c1: Currency
        q1: Decimal
        d1: Date
        c2: Currency
        q2: Decimal
        d2: Date
        c1, q1, d1 = self
        c2, q2, d2 = other  # type: ignore

        if c1 != c2:
            raise IncompatibleCurrencyError(ccy1=c1, ccy2=c2, operation="subtraction")

        return SomeMoney(c1, q1 - q2, d1 if d1 > d2 else d2)

    def scalar_subtract(self, other: Numeric) -> "Money":
        ## TODO: **try** not casting other to Decimal.
        c, q, d = self
        return SomeMoney(c, (q - Decimal(other)).quantize(c.quantizer), d)

    def multiply(self, other: Numeric) -> "Money":
        ## TODO: **try** not casting other to Decimal.
        c, q, d = self
        return SomeMoney(c, (q * Decimal(other)).quantize(c.quantizer), d)

    def divide(self, other: Numeric) -> "Money":
        ## TODO: **try** not casting other to Decimal.
        try:
            c, q, d = self
            return SomeMoney(c, (q / Decimal(other)).quantize(c.quantizer), d)
        except (InvalidOperation, DivisionByZero):
            return NoMoney

    def floor_divide(self, other: Numeric) -> "Money":
        ## TODO: **try** not casting other to Decimal.
        try:
            c, q, d = self
            return SomeMoney(c, (q // Decimal(other)).quantize(c.quantizer), d)
        except (InvalidOperation, DivisionByZero):
            return NoMoney

    def lt(self, other: "Money") -> bool:
        if other.undefined:
            return False
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation="< comparision")
        return self.qty < other.qty

    def lte(self, other: "Money") -> bool:
        if other.undefined:
            return False
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation="<= comparision")
        return self.qty <= other.qty

    def gt(self, other: "Money") -> bool:
        if other.undefined:
            return True
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation="> comparision")
        return self.qty > other.qty

    def gte(self, other: "Money") -> bool:
        if other.undefined:
            return True
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation=">= comparision")
        return self.qty >= other.qty

    def with_ccy(self, ccy: Currency) -> "Money":
        return SomeMoney(ccy, self[1], self[2])

    def with_qty(self, qty: Decimal) -> "Money":
        c, q, d = self
        return SomeMoney(c, qty.quantize(c.quantizer), d)

    def with_dov(self, dov: Date) -> "Money":
        return SomeMoney(self[0], self[1], dov)

    def ccy_or(self, default: Currency) -> Currency:
        return self[0]

    def qty_or(self, default: Decimal) -> Decimal:
        return self[1]

    def dov_or(self, default: Date) -> Date:
        return self[2]

    def convert(self, to: Currency, asof: Optional[Date] = None, strict: bool = False) -> "Money":
        ## Get slots:
        ccy, qty, dov = self

        ## Get date of conversion:
        asof = asof or dov

        ## Attempt to get the FX rate:
        try:
            rate = FXRateService.default.query(ccy, to, asof, strict)  # type: ignore
        except AttributeError as exc:
            if FXRateService.default is None:
                raise ProgrammingError("Did you implement and set the default FX rate service?")
            else:
                raise exc

        ## Do we have a rate?
        if rate is None:
            ## Nope, shall we raise exception?
            if strict:
                ## Yep:
                raise FXRateLookupError(ccy, to, asof)
            else:
                ## Just return NA:
                return NoMoney

        ## Compute and return:
        return SomeMoney(to, (qty * rate.value).quantize(to.quantizer), asof)

    @property
    def price(self) -> "Price":
        return SomePrice(*self)

    __bool__ = as_boolean

    __eq__ = is_equal

    __abs__ = abs

    __float__ = as_float

    __int__ = as_integer

    __neg__ = negative

    __pos__ = positive

    __add__ = add  # type: ignore

    __sub__ = subtract

    __mul__ = multiply  # type: ignore

    __truediv__ = divide

    __floordiv__ = floor_divide

    __lt__ = lt  # type: ignore

    __le__ = lte  # type: ignore

    __gt__ = gt  # type: ignore

    __ge__ = gte  # type: ignore


class NoneMoney(Money):

    __slots__ = ()

    defined = False

    undefined = True

    def as_boolean(self) -> bool:
        return False

    def is_equal(self, other: Any) -> bool:
        return other.__class__ is NoneMoney

    def abs(self) -> "Money":
        return self

    def as_float(self) -> float:
        raise TypeError("Undefined monetary values do not have quantity information.")

    def as_integer(self) -> int:
        raise TypeError("Undefined monetary values do not have quantity information.")

    def round(self, ndigits: int = 0) -> "Money":
        return self

    def negative(self) -> "Money":
        return self

    def positive(self) -> "Money":
        return self

    def add(self, other: "Money") -> "Money":
        return other

    def scalar_add(self, other: Numeric) -> "Money":
        return self

    def subtract(self, other: "Money") -> "Money":
        return -other

    def scalar_subtract(self, other: Numeric) -> "Money":
        return self

    def multiply(self, other: Numeric) -> "Money":
        return self

    def divide(self, other: Numeric) -> "Money":
        return self

    def floor_divide(self, other: Numeric) -> "Money":
        return self

    def lt(self, other: "Money") -> bool:
        return other.defined

    def lte(self, other: "Money") -> bool:
        return True

    def gt(self, other: "Money") -> bool:
        return False

    def gte(self, other: "Money") -> bool:
        return other.undefined

    def with_ccy(self, ccy: Currency) -> "Money":
        return self

    def with_qty(self, qty: Decimal) -> "Money":
        return self

    def with_dov(self, dov: Date) -> "Money":
        return self

    def ccy_or(self, default: Currency) -> Currency:
        return default

    def qty_or(self, default: Decimal) -> Decimal:
        return default

    def dov_or(self, default: Date) -> Date:
        return default

    def convert(self, to: Currency, asof: Optional[Date] = None, strict: bool = False) -> "Money":
        return self

    @property
    def price(self) -> "Price":
        return NoPrice

    __bool__ = as_boolean

    __eq__ = is_equal

    __abs__ = abs

    __float__ = as_float

    __int__ = as_integer

    __neg__ = negative

    __pos__ = positive

    __add__ = add

    __sub__ = subtract

    __mul__ = multiply

    __truediv__ = divide

    __floordiv__ = floor_divide

    __lt__ = lt

    __le__ = lte

    __gt__ = gt

    __ge__ = gte


## Define and attach undefined money singleton.
Money.NA = NoMoney = NoneMoney()


class Price:
    """
    Provides an abstract price model and its semantics.
    """

    ## No need for slots.
    __slots__ = ()

    #: Defines the *undefined* price object as a singleton.
    NA: "Price"

    #: Returns the currency of the price object, if defined.
    #:
    #: Note that a :class:`TypeError` is raised if call-site attempts to access this property of an undefined price.
    ccy: Currency

    #: Returns the quantity of the price object, if defined.
    #:
    #: Note that a :class:`TypeError` is raised if call-site attempts to access this property of an undefined price.
    qty: Decimal

    #: Returns the value date of the price object, if defined.
    #:
    #: Note that a :class:`TypeError` is raised if call-site attempts to access this property of an undefined price.
    dov: Date

    #: Indicates that the price is a *defined* monetary value.
    defined: bool  # noqa: E704

    #: Indicates that the price is an *undefined* monetary value.
    undefined: bool

    @abstractmethod
    def is_equal(self, other: Any) -> bool:
        """
        Checks the equality of two price objects.

        In particular:

        1. ``True`` if ``other`` is a price object **and** all slots are same.
        2. ``False`` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def as_boolean(self) -> bool:
        """
        Returns the logical representation of the price object.

        In particular:

        1. ``False`` if price is *undefined* **or** price quantity is ``zero``.
        2. ``True`` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def as_float(self) -> float:
        """
        Returns the quantity as a ``float`` if *defined*, raises class:`MonetaryOperationException` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def as_integer(self) -> int:
        """
        Returns the quantity as an ``int`` if *defined*, raises class:`MonetaryOperationException` otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def abs(self) -> "Price":
        """
        Returns the absolute price if *defined*, itself otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def negative(self) -> "Price":
        """
        Negates the quantity of the monetary value if *defined*, itself otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def positive(self) -> "Price":
        """
        Returns same monetary value if *defined*, itself otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def round(self, ndigits: int = 0) -> "Price":
        """
        Rounds the quantity of the monetary value to ``ndigits`` by using ``HALF_EVEN`` method if *defined*, itself
        otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def add(self, other: "Price") -> "Price":
        """
        Performs monetary addition on the price object and the given ``other`` price object.

        Note that::

        1. Raises :class:`IncompatibleCurrencyError` if currencies do not match.
        2. If any of the operands are undefined, returns the other one conveniently.
        3. Dates are carried forward as a result of addition of two defined price objects.
        """
        raise NotImplementedError

    @abstractmethod
    def scalar_add(self, other: Numeric) -> "Price":
        """
        Performs scalar addition on the quantity of the price.

        Note that undefined price object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def subtract(self, other: "Price") -> "Price":
        """
        Performs monetary subtraction on the price object and the given ``other`` price object.

        Note that::

        1. Raises :class:`IncompatibleCurrencyError` if currencies do not match.
        2. If any of the operands are undefined, returns the other one conveniently.
        3. Dates are carried forward as a result of addition of two defined price objects.
        """
        raise NotImplementedError

    @abstractmethod
    def scalar_subtract(self, other: Numeric) -> "Price":
        """
        Performs scalar subtraction on the quantity of the price.

        Note that undefined price object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def multiply(self, other: Numeric) -> "Price":
        """
        Performs scalar multiplication.

        Note that undefined price object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def times(self, other: Numeric) -> "Money":
        """
        Performs monetary multiplication operation.

        Note that undefined price object is returned as is.
        """
        raise NotImplementedError

    @abstractmethod
    def divide(self, other: Numeric) -> "Price":
        """
        Performs ordinary division on the price object if *defined*, itself otherwise.

        Note that division by zero yields an undefined price object.
        """
        raise NotImplementedError

    @abstractmethod
    def floor_divide(self, other: Numeric) -> "Price":
        """
        Performs floor division on the price object if *defined*, itself otherwise.

        Note that division by zero yields an undefined price object.

        """
        raise NotImplementedError

    @abstractmethod
    def lt(self, other: "Price") -> bool:
        """
        Applies "less than" comparison against ``other`` price.

        Note that::

        1. Undefined price objects are always less than ``other`` if ``other`` is not undefined, and
        2. :class:`IncompatibleCurrencyError` is raised when comparing two defined price objects with different
        currencies.
        """
        pass

    @abstractmethod
    def lte(self, other: "Price") -> bool:
        """
        Applies "less than or equal to" comparison against ``other`` price.

        Note that::

        1. Undefined price objects are always less than or equal to ``other``, and
        2. :class:`IncompatibleCurrencyError` is raised when comparing two defined price objects with different
        currencies.
        """
        pass

    @abstractmethod
    def gt(self, other: "Price") -> bool:
        """
        Applies "greater than" comparison against ``other`` price.

        Note that::

        1. Undefined price objects are never greater than ``other``,
        2. Defined price objects are always greater than ``other`` if other is undefined, and
        3. :class:`IncompatibleCurrencyError` is raised when comparing two defined price objects with different
        currencies.
        """
        pass

    @abstractmethod
    def gte(self, other: "Price") -> bool:
        """
        Applies "greater than or equal to" comparison against ``other`` price.

        Note that::

        1. Undefined price objects are never greater than or equal to ``other`` if ``other`` is defined,
        2. Undefined price objects are greater than or equal to ``other`` if ``other is undefined, and
        3. :class:`IncompatibleCurrencyError` is raised when comparing two defined price objects with different
        currencies.
        """
        pass

    @abstractmethod
    def with_ccy(self, ccy: Currency) -> "Price":
        """
        Creates a new price object with the given currency if price is *defined*, returns itself otherwise.
        """
        pass

    @abstractmethod
    def with_qty(self, qty: Decimal) -> "Price":
        """
        Creates a new price object with the given quantity if price is *defined*, returns itself otherwise.
        """
        pass

    @abstractmethod
    def with_dov(self, dov: Date) -> "Price":
        """
        Creates a new price object with the given value date if price is *defined*, returns itself otherwise.
        """
        pass

    @abstractmethod
    def ccy_or(self, default: Currency) -> Currency:
        """
        Returns the ``ccy`` if the monetary value is *defined*, ``default`` otherwise.

        >>> from pypara.currencies import Currencies
        >>> someprice = Price.of(Currencies["USD"], Decimal('1'), Date(2019, 1, 1))
        >>> someprice.ccy_or(Currencies["EUR"]).code
        'USD'
        >>> someprice = Price.of(Currencies["USD"], None, None)
        >>> someprice.ccy_or(Currencies["EUR"]).code
        'EUR'
        """
        pass

    @abstractmethod
    def qty_or(self, default: Decimal) -> Decimal:
        """
        Returns the ``qty`` if the monetary value is *defined*, ``default`` otherwise.

        >>> from pypara.currencies import Currencies
        >>> someprice = Price.of(Currencies["USD"], Decimal('1'), Date(2019, 1, 1))
        >>> someprice.qty_or(Decimal(0))
        Decimal('1')
        >>> noneprice = Price.of(None, Decimal('1'), None)
        >>> noneprice.qty_or(Decimal(0))
        Decimal('0')
        """
        pass

    @abstractmethod
    def dov_or(self, default: Date) -> Date:
        """
        Returns the ``dov`` if the monetary value is *defined*, ``default`` otherwise.

        >>> from pypara.currencies import Currencies
        >>> someprice = Price.of(Currencies["USD"], Decimal('1'), Date(2019, 1, 1))
        >>> someprice.dov_or(Date(2001, 1, 1))
        datetime.date(2019, 1, 1)
        >>> noneprice = Price.of(None, None, Date(2019, 1, 1))
        >>> noneprice.dov_or(Date(2001, 1, 1))
        datetime.date(2001, 1, 1)
        """
        pass

    @abstractmethod
    def convert(self, to: Currency, asof: Optional[Date] = None, strict: bool = False) -> "Price":
        """
        Converts the monetary value from one currency to another.

        Raises :class:`FXRateLookupError` if no foreign exchange rate can be found for conversion.

        Note that we will carry the date forward as per ``asof`` date.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def money(self) -> Money:
        """
        Returns the money representation of the price object.
        """
        raise NotImplementedError

    @classmethod
    def of(cls, ccy: Optional[Currency], qty: Optional[Decimal], dov: Optional[Date]) -> "Price":
        """
        Provides a factory method to create a new price object in a safe manner.
        """
        if qty is None or ccy is None or dov is None:
            return NoPrice
        return SomePrice(ccy, qty, dov)

    @abstractmethod
    def __bool__(self) -> bool:
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __abs__(self) -> "Price":
        pass

    @abstractmethod
    def __float__(self) -> float:
        pass

    @abstractmethod
    def __int__(self) -> int:
        pass

    @overload
    def __round__(self) -> int:
        ...

    @overload
    def __round__(self, ndigits: None) -> int:
        ...

    @overload
    def __round__(self, ndigits: int) -> "Price":
        ...

    def __round__(self, ndigits: Optional[int] = 0) -> Union["Price", int]:
        return self.round(ndigits or 0)

    @abstractmethod
    def __neg__(self) -> "Price":
        pass

    @abstractmethod
    def __pos__(self) -> "Price":
        pass

    @abstractmethod
    def __add__(self, other: "Price") -> "Price":
        pass

    @abstractmethod
    def __sub__(self, other: "Price") -> "Price":
        pass

    @abstractmethod
    def __mul__(self, other: Numeric) -> "Price":
        pass

    @abstractmethod
    def __truediv__(self, other: Numeric) -> "Price":
        pass

    @abstractmethod
    def __floordiv__(self, other: Numeric) -> "Price":
        pass

    @abstractmethod
    def __lt__(self, other: "Price") -> bool:
        pass

    @abstractmethod
    def __le__(self, other: "Price") -> bool:
        pass

    @abstractmethod
    def __gt__(self, other: "Price") -> bool:
        pass

    @abstractmethod
    def __ge__(self, other: "Price") -> bool:
        pass


class SomePrice(Price, NamedTuple("SomePrice", [("ccy", Currency), ("qty", Decimal), ("dov", Date)])):
    """
    Provides a *defined* price object model.
    """

    __slots__ = ()

    defined = True

    undefined = False

    def is_equal(self, other: Any) -> bool:
        return other.__class__ is SomePrice and tuple(self) == tuple(other)

    def as_boolean(self) -> bool:
        return self.qty.__bool__()

    def as_float(self) -> float:
        return self.qty.__float__()

    def as_integer(self) -> int:
        return self.qty.__int__()

    def abs(self) -> "Price":
        c, q, d = self
        return SomePrice(c, q.__abs__(), d)

    def negative(self) -> "Price":
        c, q, d = self
        return SomePrice(c, q.__neg__(), d)

    def positive(self) -> "Price":
        c, q, d = self
        return SomePrice(c, q.__pos__(), d)

    def round(self, ndigits: int = 0) -> "Price":
        c, q, d = self
        return SomePrice(c, q.__round__(ndigits), d)

    def add(self, other: "Price") -> "Price":
        if other.undefined:
            return self

        c1: Currency
        q1: Decimal
        d1: Date
        c2: Currency
        q2: Decimal
        d2: Date
        c1, q1, d1 = self
        c2, q2, d2 = other  # type: ignore

        if c1 != c2:
            raise IncompatibleCurrencyError(ccy1=c1, ccy2=c2, operation="addition")

        return SomePrice(c1, q1 + q2, d1 if d1 > d2 else d2)

    def scalar_add(self, other: Numeric) -> "Price":
        ## TODO: **try** not casting other to Decimal.
        c, q, d = self
        return SomePrice(c, q + Decimal(other), d)

    def subtract(self, other: "Price") -> "Price":
        if other.undefined:
            return self

        c1: Currency
        q1: Decimal
        d1: Date
        c2: Currency
        q2: Decimal
        d2: Date
        c1, q1, d1 = self
        c2, q2, d2 = other  # type: ignore

        if c1 != c2:
            raise IncompatibleCurrencyError(ccy1=c1, ccy2=c2, operation="subtraction")

        return SomePrice(c1, q1 - q2, d1 if d1 > d2 else d2)

    def scalar_subtract(self, other: Numeric) -> "Price":
        ## TODO: **try** not casting other to Decimal.
        c, q, d = self
        return SomePrice(c, q - Decimal(other), d)

    def multiply(self, other: Numeric) -> "Price":
        ## TODO: **try** not casting other to Decimal.
        c, q, d = self
        return SomePrice(c, q * Decimal(other), d)

    def times(self, other: Numeric) -> "Money":
        c, q, d = self
        return SomeMoney(c, (q * Decimal(other)).quantize(c.quantizer), self.dov)

    def divide(self, other: Numeric) -> "Price":
        ## TODO: **try** not casting other to Decimal.
        try:
            c, q, d = self
            return SomePrice(c, q / Decimal(other), d)
        except (InvalidOperation, DivisionByZero):
            return NoPrice

    def floor_divide(self, other: Numeric) -> "Price":
        ## TODO: **try** not casting other to Decimal.
        try:
            c, q, d = self
            return SomePrice(c, q // Decimal(other), d)
        except (InvalidOperation, DivisionByZero):
            return NoPrice

    def lt(self, other: "Price") -> bool:
        if other.undefined:
            return False
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation="< comparision")
        return self.qty < other.qty

    def lte(self, other: "Price") -> bool:
        if other.undefined:
            return False
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation="<= comparision")
        return self.qty <= other.qty

    def gt(self, other: "Price") -> bool:
        if other.undefined:
            return True
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation="> comparision")
        return self.qty > other.qty

    def gte(self, other: "Price") -> bool:
        if other.undefined:
            return True
        elif self.ccy != other.ccy:
            raise IncompatibleCurrencyError(ccy1=self.ccy, ccy2=other.ccy, operation=">= comparision")
        return self.qty >= other.qty

    def with_ccy(self, ccy: Currency) -> "Price":
        return SomePrice(ccy, self[1], self[2])

    def with_qty(self, qty: Decimal) -> "Price":
        return SomePrice(self[0], qty, self[2])

    def with_dov(self, dov: Date) -> "Price":
        return SomePrice(self[0], self[1], dov)

    def ccy_or(self, default: Currency) -> Currency:
        return self[0]

    def qty_or(self, default: Decimal) -> Decimal:
        return self[1]

    def dov_or(self, default: Date) -> Date:
        return self[2]

    def convert(self, to: Currency, asof: Optional[Date] = None, strict: bool = False) -> "Price":
        ## Get slots:
        ccy, qty, dov = self

        ## Get date of conversion:
        asof = asof or dov

        ## Attempt to get the FX rate:
        try:
            rate = FXRateService.default.query(ccy, to, asof, strict)  # type: ignore
        except AttributeError as exc:
            if FXRateService.default is None:
                raise ProgrammingError("Did you implement and set the default FX rate service?")
            else:
                raise exc

        ## Do we have a rate?
        if rate is None:
            ## Nope, shall we raise exception?
            if strict:
                ## Yep:
                raise FXRateLookupError(ccy, to, asof)
            else:
                ## Just return NA:
                return NoPrice

        ## Compute and return:
        return SomePrice(to, qty * rate.value, asof)

    @property
    def money(self) -> Money:
        c, q, d = self
        return SomeMoney(c, q.quantize(c.quantizer), d)

    __bool__ = as_boolean

    __eq__ = is_equal

    __abs__ = abs

    __float__ = as_float

    __int__ = as_integer

    __neg__ = negative

    __pos__ = positive

    __add__ = add  # type: ignore

    __sub__ = subtract

    __mul__ = multiply  # type: ignore

    __truediv__ = divide

    __floordiv__ = floor_divide

    __lt__ = lt  # type: ignore

    __le__ = lte  # type: ignore

    __gt__ = gt  # type: ignore

    __ge__ = gte  # type: ignore


class NonePrice(Price):

    __slots__ = ()

    defined = False

    undefined = True

    def as_boolean(self) -> bool:
        return False

    def is_equal(self, other: Any) -> bool:
        return other.__class__ is NonePrice

    def abs(self) -> "Price":
        return self

    def as_float(self) -> float:
        raise TypeError("Undefined monetary values do not have quantity information.")

    def as_integer(self) -> int:
        raise TypeError("Undefined monetary values do not have quantity information.")

    def round(self, ndigits: int = 0) -> "Price":
        return self

    def negative(self) -> "Price":
        return self

    def positive(self) -> "Price":
        return self

    def add(self, other: "Price") -> "Price":
        return other

    def scalar_add(self, other: Numeric) -> "Price":
        return self

    def subtract(self, other: "Price") -> "Price":
        return -other

    def scalar_subtract(self, other: Numeric) -> "Price":
        return self

    def multiply(self, other: Numeric) -> "Price":
        return self

    def times(self, other: Numeric) -> "Money":
        return NoMoney

    def divide(self, other: Numeric) -> "Price":
        return self

    def floor_divide(self, other: Numeric) -> "Price":
        return self

    def lt(self, other: "Price") -> bool:
        return other.defined

    def lte(self, other: "Price") -> bool:
        return True

    def gt(self, other: "Price") -> bool:
        return False

    def gte(self, other: "Price") -> bool:
        return other.undefined

    def with_ccy(self, ccy: Currency) -> "Price":
        return self

    def with_qty(self, qty: Decimal) -> "Price":
        return self

    def with_dov(self, dov: Date) -> "Price":
        return self

    def ccy_or(self, default: Currency) -> Currency:
        return default

    def qty_or(self, default: Decimal) -> Decimal:
        return default

    def dov_or(self, default: Date) -> Date:
        return default

    def convert(self, to: Currency, asof: Optional[Date] = None, strict: bool = False) -> "Price":
        return self

    money = NoMoney

    __bool__ = as_boolean

    __eq__ = is_equal

    __abs__ = abs

    __float__ = as_float

    __int__ = as_integer

    __neg__ = negative

    __pos__ = positive

    __add__ = add

    __sub__ = subtract

    __mul__ = multiply

    __truediv__ = divide

    __floordiv__ = floor_divide

    __lt__ = lt

    __le__ = lte

    __gt__ = gt

    __ge__ = gte


## Define and attach undefined price singleton.
Price.NA = NoPrice = NonePrice()



"""
This module defines common date/time typings and functions.
"""

__all__ = [
    "Date",
    "DateRange",
    "DateTime",
    "ensure_date",
    "ensure_datetime",
    "FinancialPeriods",
    "get_month_end",
    "get_month_start",
    "get_now",
    "get_period_starts",
    "get_prev_weekday",
    "get_prev_year_end",
    "get_quarter_end_stream",
    "get_quarter_end",
    "get_quarter_start",
    "get_today",
    "get_tomorrow",
    "get_week_end",
    "get_week_start",
    "get_year_end",
    "get_year_half_end",
    "get_year_half_start",
    "get_year_start",
    "get_yesterday",
    "make_financial_periods",
    "OpenDateRange",
    "PeriodStarts",
    "Time",
    "TimeDelta",
]

from calendar import isleap
from dataclasses import dataclass
from datetime import date as Date
from datetime import datetime as DateTime
from datetime import time as Time
from datetime import timedelta as TimeDelta
from typing import Dict, Iterable, Iterator, Literal, Optional, OrderedDict, Tuple, Union

from dateutil.parser import ParserError, parse
from dateutil.relativedelta import relativedelta

from pypara.commons.numbers import NaturalNumber, PositiveInteger


class OpenDateRange:
    """
    Defines a daterange class with inclusive but optional start and end.

    For a date range with both ends open:

    >>> range = OpenDateRange()
    >>> range.start is None
    True
    >>> range.end is None
    True
    >>> range.is_finite
    False
    >>> range.has_start
    False
    >>> range.has_end
    False
    >>> list(range.range)
    []
    >>> range.endpoints
    (None, None)

    For a date range with both ends closed:

    >>> range2 = OpenDateRange(Date(2018, 1, 1), Date(2018, 1, 3))
    >>> range2.start is None
    False
    >>> range2.end is None
    False
    >>> range2.is_finite
    True
    >>> range2.has_start
    True
    >>> range2.has_end
    True
    >>> list(range2.range)
    [datetime.date(2018, 1, 1), datetime.date(2018, 1, 2), datetime.date(2018, 1, 3)]
    >>> range2.endpoints
    (datetime.date(2018, 1, 1), datetime.date(2018, 1, 3))
    """

    def __init__(self, start: Optional[Date] = None, end: Optional[Date] = None) -> None:
        ## Cast the start date:
        self.start = start

        ## Cast the end date:
        self.end = end

    @property
    def is_finite(self) -> bool:
        return self.start is not None and self.end is not None

    @property
    def has_start(self) -> bool:
        return self.start is not None

    @property
    def has_end(self) -> bool:
        return self.end is not None

    @property
    def range(self) -> Iterator[Date]:
        if self.start is None or self.end is None:
            return iter(())
        return self._drange(self.start, self.end)

    @property
    def endpoints(self) -> Tuple[Optional[Date], Optional[Date]]:
        return self.start, self.end

    @staticmethod
    def _drange(start: Date, end: Date) -> Iterator[Date]:
        """
        Returns a date range.
        """
        while start <= end:
            yield start
            start = start + TimeDelta(days=1)


@dataclass(frozen=True)
class DateRange(Iterable[Date]):
    """
    Provides an encoding for date ranges with inclusive date endpoints.

    A :py:class:`DateRange` is essentially a simple value object with two
    fields: :py:attr:`since` and :py:attr:`until`. These denote the start and
    end dates of the date range.

    Once successfully initialized, the value is guaranteed to have the property
    of :py:attr:`until` being equal to or greater than :py:attr:`since`.

    In an attempt to create a :py:class:`DateRange` instance that violates this
    condition, an :py:class:`AssertionError` is raised. In other words, the
    constructor of :py:meth:`DateRange.__init__` is unsafe. The safe alternative
    of creating :py:class:`DateRange` instances is :py:meth:`DateRange.of`. It
    returns a :py:class:`DateRange` instance if the condition is satisfied,
    ``None`` otherwise.

    Below are some examples of successful creation and erroneous attempt of
    creation:

    >>> good = DateRange(Date(2019, 1, 1), Date(2019, 1, 1))
    >>> good.since
    datetime.date(2019, 1, 1)
    >>> good.until
    datetime.date(2019, 1, 1)
    >>> list(good)
    [datetime.date(2019, 1, 1)]
    >>> better = DateRange(Date(2019, 1, 1), Date(2019, 1, 2))
    >>> better.since
    datetime.date(2019, 1, 1)
    >>> better.until
    datetime.date(2019, 1, 2)
    >>> list(better)
    [datetime.date(2019, 1, 1), datetime.date(2019, 1, 2)]
    >>> best = DateRange(Date(2019, 1, 1), Date(2019, 1, 3))
    >>> best.since
    datetime.date(2019, 1, 1)
    >>> best.until
    datetime.date(2019, 1, 3)
    >>> list(best)
    [datetime.date(2019, 1, 1), datetime.date(2019, 1, 2), datetime.date(2019, 1, 3)]

    In case of bad input:

    >>> bad = DateRange(Date(2019, 1, 2), Date(2019, 1, 1))
    Traceback (most recent call last):
    ...
        assert self.since > self.until
    AssertionError

    Therefore, it is advised to use :py:meth:`DateRange.of` instead:

    >>> DateRange.of(Date(2019, 1, 2), Date(2019, 1, 1)) is None
    True
    >>> DateRange.of(Date(2019, 1, 1), Date(2019, 1, 1))
    DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 1))
    """

    #: Date the range starts from (inclusive).
    since: Date

    #: Date the range ends on (inclusive).
    until: Date

    def __post_init__(self) -> None:
        """
        Checks whether the two date endpoints are consistent.

        :raises AssertError: If ``since`` is later than ``until``.
        """
        assert self.since <= self.until

    def __iter__(self) -> Iterator[Date]:
        """
        Returns an iterator for dates within the date-range in ascending order.

        :return: An :py:class:`typing.Iterator` of :py:class:`datetime.date` instances.
        """
        return iter((self.since + TimeDelta(days=i) for i in range(0, (self.until - self.since).days + 1)))

    @classmethod
    def of(cls, since: Date, until: Date) -> Optional["DateRange"]:
        """
        Safe, smart constructor to create :py:class:`DateRange` instances.

        :param since: Start date of the date range.
        :param until: End date of the date range.
        :return: A :py:class:`DateRange` instance if the ``since <= until`` is satisfied, ``None`` otherwise.

        >>> DateRange.of(Date(2019, 1, 2), Date(2019, 1, 1))
        >>> DateRange.of(Date(2019, 1, 1), Date(2019, 1, 1))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 1))
        """
        return cls(since, until) if since <= until else None

    @classmethod
    def ordered(cls, date1: Date, date2: Date) -> "DateRange":
        """
        Safe, smart constructor to create :py:class:`DateRange` instances ensuring the order of dates.

        :param date1: First date.
        :param date2: Second date.
        :return: A :py:class:`DateRange` instance with ``(since, until)`` as ``(date1, date2)`` if ``date1 <= date2``
                 ``(date2, date1)`` otherwise.

        >>> DateRange.ordered(Date(2019, 1, 1), Date(2019, 1, 1))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 1))
        >>> DateRange.ordered(Date(2019, 1, 1), Date(2019, 1, 2))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 2))
        >>> DateRange.ordered(Date(2019, 1, 2), Date(2019, 1, 1))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 2))
        """
        return cls(date1, date2) if date1 <= date2 else cls(date2, date1)

    @classmethod
    def pivotal(cls, pivot: Date, prev: NaturalNumber, next: NaturalNumber) -> "DateRange":
        """
        Creates a :py:class:`DateRange` as per given ``pivot`` date in time to create date range since ``prev`` years
        back and ``next`` years forward.

        :param pivot: Pivot date in time.
        :param prev: Numbers to go back in years.
        :param next: Numbers to go forward in years.
        :return: A :py:class:`DateRange` instance.

        >>> DateRange.pivotal(Date(2019, 12, 31), NaturalNumber(0), NaturalNumber(0))
        DateRange(since=datetime.date(2019, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange.pivotal(Date(2019, 12, 31), NaturalNumber(0), NaturalNumber(1))
        DateRange(since=datetime.date(2019, 12, 31), until=datetime.date(2020, 12, 31))
        >>> DateRange.pivotal(Date(2019, 12, 31), NaturalNumber(1), NaturalNumber(0))
        DateRange(since=datetime.date(2018, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange.pivotal(Date(2019, 12, 31), NaturalNumber(1), NaturalNumber(1))
        DateRange(since=datetime.date(2018, 12, 31), until=datetime.date(2020, 12, 31))
        >>> DateRange.pivotal(Date(2020, 2, 29), NaturalNumber(0), NaturalNumber(0))
        DateRange(since=datetime.date(2020, 2, 29), until=datetime.date(2020, 2, 29))
        >>> DateRange.pivotal(Date(2020, 2, 29), NaturalNumber(0), NaturalNumber(1))
        DateRange(since=datetime.date(2020, 2, 29), until=datetime.date(2021, 2, 28))
        >>> DateRange.pivotal(Date(2020, 2, 29), NaturalNumber(1), NaturalNumber(0))
        DateRange(since=datetime.date(2019, 2, 28), until=datetime.date(2020, 2, 29))
        >>> DateRange.pivotal(Date(2020, 2, 29), NaturalNumber(1), NaturalNumber(1))
        DateRange(since=datetime.date(2019, 2, 28), until=datetime.date(2021, 2, 28))
        """
        ## Get the target since date:
        sy, sm, sd = pivot.year - prev, pivot.month, pivot.day

        ## Check if (m, d) is a valid one:
        if not isleap(sy) and sm == 2 and sd == 29:
            sd = 28

        ## Get the target until date:
        uy, um, ud = pivot.year + next, pivot.month, pivot.day

        ## Check if (m, d) is a valid one:
        if not isleap(uy) and um == 2 and ud == 29:
            ud = 28

        ## Create the date range and return:
        return cls(Date(sy, sm, sd), Date(uy, um, ud))

    @classmethod
    def dtd(cls, date: Date) -> "DateRange":
        """
        Returns day-to-date date range as of the given date.

        :param date: To-date.
        :return: Day-to-date (DTD) date range.

        >>> DateRange.dtd(Date(2019, 1, 9))
        DateRange(since=datetime.date(2019, 1, 9), until=datetime.date(2019, 1, 9))
        >>> DateRange.dtd(Date(2020, 3, 1))
        DateRange(since=datetime.date(2020, 3, 1), until=datetime.date(2020, 3, 1))
        >>> DateRange.dtd(Date(2020, 3, 9))
        DateRange(since=datetime.date(2020, 3, 9), until=datetime.date(2020, 3, 9))
        """
        return cls(date, date)

    @classmethod
    def mtd(cls, date: Date) -> "DateRange":
        """
        Returns month-to-date date range as of the given date.

        :param date: To-date.
        :return: Month-to-date (MTD) date range.

        >>> DateRange.mtd(Date(2019, 1, 9))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 9))
        >>> DateRange.mtd(Date(2020, 3, 1))
        DateRange(since=datetime.date(2020, 3, 1), until=datetime.date(2020, 3, 1))
        >>> DateRange.mtd(Date(2020, 3, 9))
        DateRange(since=datetime.date(2020, 3, 1), until=datetime.date(2020, 3, 9))
        """
        return cls(date.replace(day=1), date)

    @classmethod
    def ytd(cls, date: Date) -> "DateRange":
        """
        Returns year-to-date date range as of the given date.

        :param date: To-date.
        :return: Year-to-date (YTD) date range.

        >>> DateRange.ytd(Date(2020, 1, 1))
        DateRange(since=datetime.date(2020, 1, 1), until=datetime.date(2020, 1, 1))
        >>> DateRange.ytd(Date(2019, 1, 9))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 1, 9))
        >>> DateRange.ytd(Date(2020, 3, 1))
        DateRange(since=datetime.date(2020, 1, 1), until=datetime.date(2020, 3, 1))
        """
        return cls(date.replace(month=1, day=1), date)

    @classmethod
    def year(cls, year: PositiveInteger) -> "DateRange":
        """
        Returns a full year date range for the given year.

        :param year: The year.
        :return: Year date range.

        >>> DateRange.year(PositiveInteger(2019))
        DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 12, 31))
        >>> DateRange.year(PositiveInteger(2020))
        DateRange(since=datetime.date(2020, 1, 1), until=datetime.date(2020, 12, 31))
        """
        return cls(Date(year, 1, 1), Date(year, 12, 31))

    @classmethod
    def cover(cls, first: "DateRange", *rest: "DateRange") -> "DateRange":
        """
        Returns the date-range that covers all given date ranges.

        :param first: First date range.
        :param rest: Rest of date ranges.
        :return: A new date range which covers all given date ranges.

        >>> DateRange.cover(DateRange.dtd(Date(2019, 4, 1)), DateRange.mtd(Date(2020, 2, 29)), DateRange.ytd(Date(2018, 3, 6)))  # noqa: E501
        DateRange(since=datetime.date(2018, 1, 1), until=datetime.date(2020, 2, 29))
        """
        return DateRange(min(first, *rest, key=lambda x: x.since).since, max(first, *rest, key=lambda x: x.until).until)

    def since_prev_year_end(self, years: PositiveInteger = PositiveInteger(1), weekday: bool = False) -> "DateRange":
        """
        Re-creates the date range starting from the previous ``years`` number of
        years as per :py:attr:`DateRange.since`.

        >>> DateRange(Date(2019, 12, 31), Date(2019, 12, 31)).since_prev_year_end(PositiveInteger(1))
        DateRange(since=datetime.date(2018, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange(Date(2019, 12, 30), Date(2019, 12, 31)).since_prev_year_end(PositiveInteger(1))
        DateRange(since=datetime.date(2018, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange(Date(2019, 12, 30), Date(2019, 12, 31)).since_prev_year_end(PositiveInteger(2))
        DateRange(since=datetime.date(2017, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange(Date(2019, 12, 31), Date(2019, 12, 31)).since_prev_year_end(PositiveInteger(1), True)
        DateRange(since=datetime.date(2018, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange(Date(2019, 12, 30), Date(2019, 12, 31)).since_prev_year_end(PositiveInteger(1), True)
        DateRange(since=datetime.date(2018, 12, 31), until=datetime.date(2019, 12, 31))
        >>> DateRange(Date(2019, 12, 30), Date(2019, 12, 31)).since_prev_year_end(PositiveInteger(2), True)
        DateRange(since=datetime.date(2017, 12, 29), until=datetime.date(2019, 12, 31))

        :param years: Number of years to go back as a positive integer.
        :param weekday: Indicates if we want the year end to be a weekday.
        :return: A new :py:class:`DateRange` instance.
        """
        ## Get the year end as requested:
        yearend = get_prev_year_end(self.since, years)

        ## Create the new date range and return:
        return DateRange(get_prev_weekday(yearend) if weekday and yearend.weekday() in {5, 6} else yearend, self.until)


#: Defines a custom financial periods container.
FinancialPeriods = Dict[str, DateRange]


def make_financial_periods(date: Date, lookback: PositiveInteger) -> FinancialPeriods:
    """
    Creates financial periods for the given date with ``lookback`` number of years.

    :param date: Date to create financial periods for.
    :param lookback: A positive integer describing the number of years to go back.
    :return: A dictionary of financial period label and date range for the financual period.

    >>> sorted(make_financial_periods(Date(2019, 1, 10), PositiveInteger(1)).keys())
    ['2018', 'DTD', 'MTD', 'YTD']
    >>> sorted(make_financial_periods(Date(2019, 1, 10), PositiveInteger(2)).keys())
    ['2017', '2018', 'DTD', 'MTD', 'YTD']
    >>> sorted(make_financial_periods(Date(2019, 1, 10), PositiveInteger(3)).keys())
    ['2016', '2017', '2018', 'DTD', 'MTD', 'YTD']
    >>> sorted(make_financial_periods(Date(2020, 2, 29), PositiveInteger(3)).keys())
    ['2017', '2018', '2019', 'DTD', 'MTD', 'YTD']
    >>> periods = make_financial_periods(Date(2020, 2, 29), PositiveInteger(3))
    >>> periods["DTD"]
    DateRange(since=datetime.date(2020, 2, 29), until=datetime.date(2020, 2, 29))
    >>> periods["MTD"]
    DateRange(since=datetime.date(2020, 2, 1), until=datetime.date(2020, 2, 29))
    >>> periods["YTD"]
    DateRange(since=datetime.date(2020, 1, 1), until=datetime.date(2020, 2, 29))
    >>> periods["2019"]
    DateRange(since=datetime.date(2019, 1, 1), until=datetime.date(2019, 12, 31))
    >>> periods["2018"]
    DateRange(since=datetime.date(2018, 1, 1), until=datetime.date(2018, 12, 31))
    >>> periods["2017"]
    DateRange(since=datetime.date(2017, 1, 1), until=datetime.date(2017, 12, 31))
    """
    ## Get years iterable:
    years = (date.year - i for i in range(1, lookback + 1) if i <= date.year)

    ## Build ranges and return:
    return OrderedDict(
        (
            ("DTD", DateRange.dtd(date)),
            ("MTD", DateRange.mtd(date)),
            ("YTD", DateRange.ytd(date)),
            *((f"{y}", DateRange.year(PositiveInteger(y))) for y in years),
        )
    )


def get_now(**kwargs: int) -> DateTime:
    """
    Returns the current date/time with replacements if provided.

    >>> get_now(year=1981, month=8, day=27, hour=18, minute=19, second=20, microsecond=1)
    datetime.datetime(1981, 8, 27, 18, 19, 20, 1)
    """
    return DateTime.now().replace(**kwargs)  # type: ignore


def get_today(**kwargs: int) -> Date:
    """
    Returns the current date with replacements if provided.

    >>> get_today(year=1981, month=8, day=27)
    datetime.date(1981, 8, 27)
    """
    return Date.today().replace(**kwargs)


def get_yesterday(x: Optional[Date] = None) -> Date:
    """
    Returns yesterday as of given (optional) date.

    >>> get_yesterday() == get_today() - TimeDelta(days=1)
    True
    >>> get_yesterday(Date(2019, 1, 1))
    datetime.date(2018, 12, 31)
    """
    return (x or get_today()) - TimeDelta(days=1)


def get_tomorrow(x: Optional[Date] = None) -> Date:
    """
    Returns tomorrow as of given (optional) date.

    >>> get_tomorrow() == get_today() + TimeDelta(days=1)
    True
    >>> get_tomorrow(Date(2018, 12, 31))
    datetime.date(2019, 1, 1)
    """
    return (x or get_today()) + TimeDelta(days=1)


def get_prev_weekday(x: Optional[Date] = None) -> Date:
    """
    Returns the previous week day as of given (optional) date.

    :param x: Optional date in time.
    :return: Previous business day.

    >>> get_prev_weekday(Date(2020, 1, 1))
    datetime.date(2019, 12, 31)
    >>> get_prev_weekday(Date(2020, 1, 2))
    datetime.date(2020, 1, 1)
    >>> get_prev_weekday(Date(2020, 1, 3))
    datetime.date(2020, 1, 2)
    >>> get_prev_weekday(Date(2020, 1, 4))
    datetime.date(2020, 1, 3)
    >>> get_prev_weekday(Date(2020, 1, 5))
    datetime.date(2020, 1, 3)
    >>> get_prev_weekday(Date(2020, 1, 6))
    datetime.date(2020, 1, 3)
    >>> get_prev_weekday(Date(2020, 1, 7))
    datetime.date(2020, 1, 6)
    """
    ## Get the day:
    x = x or get_today()

    ## Define the offset:
    offset = max(1, (x.weekday() + 6) % 7 - 3)

    ## Compute the day and return:
    return x - TimeDelta(days=offset)


def get_prev_year_end(x: Optional[Date] = None, years: PositiveInteger = PositiveInteger(1)) -> Date:
    """
    Returns the year end of the previous year as of given (optional) date.

    >>> get_prev_year_end() == Date(get_today().year - 1, 12, 31)
    True
    >>> get_prev_year_end(Date(2019, 1, 1))
    datetime.date(2018, 12, 31)
    >>> get_prev_year_end(Date(2018, 12, 31))
    datetime.date(2017, 12, 31)
    >>> get_prev_year_end(Date(2018, 12, 31), PositiveInteger(2))
    datetime.date(2016, 12, 31)
    """
    return Date((x or get_today()).year - years, 12, 31)


def get_year_start(x: Optional[Date] = None) -> Date:
    """
    Returns the start of the year as of the given date.

    >>> get_year_start(Date(2017, 1, 1))
    datetime.date(2017, 1, 1)
    >>> get_year_start(Date(2017, 7, 29))
    datetime.date(2017, 1, 1)
    >>> get_year_start(Date(2017, 12, 31))
    datetime.date(2017, 1, 1)
    """
    return (x or get_today()).replace(month=1, day=1)


def get_year_end(x: Optional[Date] = None) -> Date:
    """
    Returns the start of the year as of the given date.

    >>> get_year_end(Date(2017, 1, 1))
    datetime.date(2017, 12, 31)
    >>> get_year_end(Date(2017, 7, 29))
    datetime.date(2017, 12, 31)
    >>> get_year_end(Date(2017, 12, 31))
    datetime.date(2017, 12, 31)
    """
    return (x or get_today()).replace(month=12, day=31)


def get_year_half_start(x: Optional[Date] = None) -> Date:
    """
    Returns the half year start as of the given date.

    >>> get_year_half_start(Date(2017, 1, 1))
    datetime.date(2017, 1, 1)
    >>> get_year_half_start(Date(2017, 5, 31))
    datetime.date(2017, 1, 1)
    >>> get_year_half_start(Date(2017, 6, 30))
    datetime.date(2017, 1, 1)
    >>> get_year_half_start(Date(2017, 7, 1))
    datetime.date(2017, 7, 1)
    >>> get_year_half_start(Date(2017, 7, 29))
    datetime.date(2017, 7, 1)
    >>> get_year_half_start(Date(2017, 12, 31))
    datetime.date(2017, 7, 1)
    """
    asof = x or get_today()
    return asof.replace(month=((asof.month - 1) // 6) * 6 + 1, day=1)


def get_year_half_end(x: Optional[Date] = None) -> Date:
    """
    Returns the half year end as of the given date.

    >>> get_year_half_end(Date(2017, 1, 1))
    datetime.date(2017, 6, 30)
    >>> get_year_half_end(Date(2017, 5, 31))
    datetime.date(2017, 6, 30)
    >>> get_year_half_end(Date(2017, 6, 30))
    datetime.date(2017, 6, 30)
    >>> get_year_half_end(Date(2017, 7, 1))
    datetime.date(2017, 12, 31)
    >>> get_year_half_end(Date(2017, 7, 29))
    datetime.date(2017, 12, 31)
    >>> get_year_half_end(Date(2017, 12, 31))
    datetime.date(2017, 12, 31)
    """
    return get_year_half_start(x or get_today()) + relativedelta(months=+6, days=-1)


def get_quarter_start(x: Optional[Date] = None) -> Date:
    """
    Returns the quarter start as of the given date.

    >>> get_quarter_start(Date(2017, 1, 1))
    datetime.date(2017, 1, 1)
    >>> get_quarter_start(Date(2017, 5, 31))
    datetime.date(2017, 4, 1)
    >>> get_quarter_start(Date(2017, 6, 30))
    datetime.date(2017, 4, 1)
    >>> get_quarter_start(Date(2017, 7, 1))
    datetime.date(2017, 7, 1)
    >>> get_quarter_start(Date(2017, 7, 29))
    datetime.date(2017, 7, 1)
    >>> get_quarter_start(Date(2017, 12, 31))
    datetime.date(2017, 10, 1)
    """
    asof = x or get_today()
    return asof.replace(month=(asof.month - 1) // 3 * 3 + 1, day=1)


def get_quarter_end(x: Optional[Date] = None) -> Date:
    """
    Returns the quarter end as of the given date.

    >>> get_quarter_end(Date(2017, 1, 1))
    datetime.date(2017, 3, 31)
    >>> get_quarter_end(Date(2017, 5, 31))
    datetime.date(2017, 6, 30)
    >>> get_quarter_end(Date(2017, 6, 30))
    datetime.date(2017, 6, 30)
    >>> get_quarter_end(Date(2017, 7, 1))
    datetime.date(2017, 9, 30)
    >>> get_quarter_end(Date(2017, 7, 29))
    datetime.date(2017, 9, 30)
    >>> get_quarter_end(Date(2017, 9, 30))
    datetime.date(2017, 9, 30)
    >>> get_quarter_end(Date(2017, 10, 1))
    datetime.date(2017, 12, 31)
    >>> get_quarter_end(Date(2017, 12, 31))
    datetime.date(2017, 12, 31)
    """
    return get_quarter_start(x or get_today()) + relativedelta(months=+3, days=-1)


def get_month_start(x: Optional[Date] = None) -> Date:
    """
    Returns the start of the month as of the given date.

    >>> get_month_start(Date(2017, 2, 1))
    datetime.date(2017, 2, 1)
    >>> get_month_start(Date(2017, 2, 2))
    datetime.date(2017, 2, 1)
    >>> get_month_start(Date(2017, 3, 31))
    datetime.date(2017, 3, 1)
    """
    return (x or get_today()).replace(day=1)


def get_month_end(x: Optional[Date] = None) -> Date:
    """
    Returns the end of the month as of the given date.

    >>> get_month_end(Date(2017, 1, 1))
    datetime.date(2017, 1, 31)
    >>> get_month_end(Date(2017, 1, 2))
    datetime.date(2017, 1, 31)
    >>> get_month_end(Date(2017, 1, 31))
    datetime.date(2017, 1, 31)
    >>> get_month_end(Date(2017, 2, 1))
    datetime.date(2017, 2, 28)
    >>> get_month_end(Date(2017, 3, 1))
    datetime.date(2017, 3, 31)
    """
    return (x or get_today()).replace(day=1) + relativedelta(months=+1, days=-1)


def get_week_start(x: Optional[Date] = None) -> Date:
    """
    Returns the start of the week as of the given date.

    >>> get_week_start(Date(2017, 1, 15))
    datetime.date(2017, 1, 9)
    >>> get_week_start(Date(2017, 1, 16))
    datetime.date(2017, 1, 16)
    >>> get_week_start(Date(2017, 1, 18))
    datetime.date(2017, 1, 16)
    """
    asof = x or get_today()
    return asof - TimeDelta(days=(asof.isoweekday() - 1) % 7)


def get_week_end(x: Optional[Date] = None) -> Date:
    """
    Returns the end of the week as of the given date.

    >>> get_week_end(Date(2017, 1, 15))
    datetime.date(2017, 1, 15)
    >>> get_week_end(Date(2017, 1, 16))
    datetime.date(2017, 1, 22)
    >>> get_week_end(Date(2017, 1, 18))
    datetime.date(2017, 1, 22)
    >>> get_week_end(Date(2017, 1, 17))
    datetime.date(2017, 1, 22)
    >>> get_week_end(Date(2017, 1, 22))
    datetime.date(2017, 1, 22)
    """
    asof = x or get_today()
    return asof + TimeDelta(days=6 - (asof.isoweekday() - 1) % 7)


#: Type encoding for a lookup table of period starts.
PeriodStarts = Dict[
    Literal["year_start", "half_start", "quarter_start", "month_start", "week_start", "yesterday"], Date
]


def get_period_starts(x: Optional[Date] = None) -> PeriodStarts:
    """
    Returns important dates as of the given date.

    >>> get_period_starts(Date(2018, 8, 19))["year_start"]
    datetime.date(2018, 1, 1)
    >>> get_period_starts(Date(2018, 8, 19))["half_start"]
    datetime.date(2018, 7, 1)
    >>> get_period_starts(Date(2018, 8, 19))["quarter_start"]
    datetime.date(2018, 7, 1)
    >>> get_period_starts(Date(2018, 8, 19))["month_start"]
    datetime.date(2018, 8, 1)
    >>> get_period_starts(Date(2018, 8, 19))["week_start"]
    datetime.date(2018, 8, 13)
    >>> get_period_starts(Date(2018, 8, 19))["yesterday"]
    datetime.date(2018, 8, 18)
    """
    asof = x or get_today()
    return OrderedDict(
        [
            ("year_start", get_year_start(asof)),
            ("half_start", get_year_half_start(asof)),
            ("quarter_start", get_quarter_start(asof)),
            ("month_start", get_month_start(asof)),
            ("week_start", get_week_start(asof)),
            ("yesterday", asof - TimeDelta(days=1)),
        ]
    )


def get_quarter_end_stream(x: Optional[Date] = None) -> Iterator[Date]:
    """
    Returns a descending stream of quarter end dates.

    >>> from itertools import islice
    >>> list(islice(get_quarter_end_stream(Date(2018, 8, 12)), 3))
    [datetime.date(2018, 6, 30), datetime.date(2018, 3, 31), datetime.date(2017, 12, 31)]
    """
    ## Get asof date:
    asof = x or get_today()

    ## Get the last quarter end:
    e = get_quarter_start(asof) - TimeDelta(days=1)

    ## Yield this:
    yield e

    ## Forever:
    while True:
        e = get_month_end(e - relativedelta(months=3))
        yield e


def ensure_datetime(value: Union[Date, DateTime, str], **kwargs: int) -> DateTime:
    """
    Attempts to convert the value to a `datetime.datetime` instance
    with the date/time fields replaced by `kwargs` if given.

    >>> ensure_datetime(Date(2015, 10, 10))
    datetime.datetime(2015, 10, 10, 0, 0)
    >>> ensure_datetime(DateTime(2015, 10, 10))
    datetime.datetime(2015, 10, 10, 0, 0)
    >>> ensure_datetime("2015-10-10")
    datetime.datetime(2015, 10, 10, 0, 0)
    >>> ensure_datetime(DateTime(2015, 10, 10), hour=12)
    datetime.datetime(2015, 10, 10, 12, 0)
    >>> ensure_datetime(DateTime(2015, 10, 10), minute=59)
    datetime.datetime(2015, 10, 10, 0, 59)
    >>> ensure_datetime("2015-10-10", second=30)
    datetime.datetime(2015, 10, 10, 0, 0, 30)
    >>> ensure_datetime("2015-10-10", second=0)
    datetime.datetime(2015, 10, 10, 0, 0)
    >>> ensure_datetime("2015-10-10", microsecond=10)
    datetime.datetime(2015, 10, 10, 0, 0, 0, 10)
    >>> ensure_datetime("A")
    Traceback (most recent call last):
    ...
    ValueError: Can not parse value into a date/time object: A
    >>> ensure_datetime(1)
    Traceback (most recent call last):
    ...
    ValueError: Don't know how to convert value to date/time object: 1
    """
    ## Check the type of the value and act accordinly.
    if isinstance(value, DateTime):
        ## It is a datetime instance. Nothing to be done. Just return with replacement:
        return value.replace(**kwargs)  # type: ignore
    elif isinstance(value, Date):
        ## It is a date instance. Set to morning and return with replacement:
        return DateTime.combine(value, DateTime.min.time()).replace(**kwargs)  # type: ignore
    elif isinstance(value, str):
        ## We have a string. Attempt to parse and return with replacement:
        try:
            return parse(value).replace(**kwargs)  # type: ignore
        except ParserError:
            raise ValueError("Can not parse value into a date/time object: {}".format(value))

    ## We have a problem here: Don't know how to convert other
    ## object. Raise a value error:
    raise ValueError("Don't know how to convert value to date/time object: {}".format(value))


def ensure_date(value: Union[Date, DateTime, str], **kwargs: int) -> Date:
    """
    Attempts to convert the value to a `datetime.date` instance with the
    date fields replaced by `kwargs` if given.

    >>> ensure_date(DateTime(2015, 10, 10))
    datetime.date(2015, 10, 10)

    >>> ensure_date(DateTime(2015, 10, 10))
    datetime.date(2015, 10, 10)

    >>> ensure_date("2015-10-10")
    datetime.date(2015, 10, 10)

    >>> ensure_date(DateTime(2015, 10, 10), hour=12)
    datetime.date(2015, 10, 10)

    >>> ensure_date(DateTime(2015, 10, 10), minute=59)
    datetime.date(2015, 10, 10)

    >>> ensure_date("2015-10-10", second=30)
    datetime.date(2015, 10, 10)

    >>> ensure_date("2015-10-10", second=0)
    datetime.date(2015, 10, 10)

    >>> ensure_date("2015-10-10", microsecond=10)
    datetime.date(2015, 10, 10)
    """
    return ensure_datetime(value, **kwargs).date()



"""
This module provides common error definitions and routines for :py:mod:`pypara`.
"""

__all__ = ["ProgrammingError"]

from typing import Optional


class ProgrammingError(Exception):
    """
    Provides a programming error exception.

    The rationale for this exception is to raise them whenever we rely on meta-programming and the programmer has
    introduced a statement which breaks the coherence of the domain logic.
    """

    @classmethod
    def passert(cls, condition: bool, message: Optional[str] = None) -> None:
        """
        Raises a :py:class:`ProgrammingError` if the condition is ``False``.

        :param condition: Indicates if the expectation is fulfilled.
        :param message: Message of the error to be raised in case that the condition is not met.
        :raises ProgrammingError: In case that the condition is ``False``.

        >>> ProgrammingError.passert(1 == 0)
        Traceback (most recent call last):
        ...
        pypara.commons.errors.ProgrammingError: Broken coherence. Check your code against domain logic to fix it.
        """
        if not condition:
            raise cls(message or "Broken coherence. Check your code against domain logic to fix it.")



"""
This module provides data definitions and functionality related to journal entries and postings.
"""

__all__ = [
    "Direction",
    "JournalEntry",
    "Posting",
    "ReadJournalEntries",
]

import datetime
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Generic, Iterable, List, Protocol, Set, TypeVar

from ..commons.guid import Guid, makeguid
from ..commons.numbers import Amount, Quantity, isum
from ..commons.zeitgeist import DateRange
from .accounts import Account, AccountType

#: Defines a type variable.
_T = TypeVar("_T")


class Direction(Enum):
    """
    Provides an enumeration for indicating increment and decrement events.
    """

    #: Declares the value type.
    value: int

    #: Indicates increment events.
    INC = +1

    #: Indicates decrement events.
    DEC = -1

    @classmethod
    def of(cls, quantity: Quantity) -> "Direction":
        """
        Returns the corresponding direction as per the sign of the quantity.

        :param quantity: Quantity to find the direction of.
        :return: Direction for the quantity.
        :raises AssertionError: If quantity is zero which implies a programming error.
        """
        assert not quantity.is_zero(), "Encountered a `0` quantity. This implies a programming error."
        return Direction.INC if quantity > 0 else Direction.DEC


#: Provides the mapping for DEBIT/CREDIT convention as per increment/decrement and account type.
_debit_mapping: Dict[Direction, Set[AccountType]] = {
    Direction.INC: {AccountType.ASSETS, AccountType.EQUITIES, AccountType.LIABILITIES},
    Direction.DEC: {AccountType.REVENUES, AccountType.EXPENSES},
}


@dataclass(frozen=True)
class Posting(Generic[_T]):
    """
    Provides a posting value object model.
    """

    #: Journal entry the posting belongs to.
    journal: "JournalEntry[_T]"

    #: Date of posting.
    date: datetime.date

    #: Account of the posting.
    account: Account

    #: Direction of the posting.
    direction: Direction

    #: Posted amount (in absolute value).
    amount: Amount

    @property
    def is_debit(self) -> bool:
        """
        Indicates if this posting is a debit.
        """
        return self.account.type in _debit_mapping[self.direction]

    @property
    def is_credit(self) -> bool:
        """
        Indicates if this posting is a credit.
        """
        return not self.is_debit


@dataclass(frozen=True)
class JournalEntry(Generic[_T]):
    """
    Provides a journal entry model.
    """

    #: Date of the entry.
    date: datetime.date

    #: Description of the entry.
    description: str

    #: Business object as the source of the journal entry.
    source: _T

    #: Postings of the journal entry.
    postings: List[Posting[_T]] = field(default_factory=list, init=False)

    #: Globally unique, ephemeral identifier.
    guid: Guid = field(default_factory=makeguid, init=False)

    @property
    def increments(self) -> Iterable[Posting[_T]]:
        """
        Incerement event postings of the journal entry.
        """
        return (p for p in self.postings if p.direction == Direction.INC)

    @property
    def decrements(self) -> Iterable[Posting[_T]]:
        """
        Decrement event postings of the journal entry.
        """
        return (p for p in self.postings if p.direction == Direction.DEC)

    @property
    def debits(self) -> Iterable[Posting[_T]]:
        """
        Debit postings of the journal entry.
        """
        return (p for p in self.postings if p.is_debit)

    @property
    def credits(self) -> Iterable[Posting[_T]]:
        """
        Credit postings of the journal entry.
        """
        return (p for p in self.postings if p.is_credit)

    def post(self, date: datetime.date, account: Account, quantity: Quantity) -> "JournalEntry[_T]":
        """
        Posts an increment/decrement event (depending on the sign of ``quantity``) to the given account.

        If the quantity is ``0``, nothing is posted.

        :param date: Date of posting.
        :param account: Account to post the amount to.
        :param quantity: Signed-value to post to the account.
        :return: This journal entry (to be chained conveniently).
        """
        if not quantity.is_zero():
            self.postings.append(Posting(self, date, account, Direction.of(quantity), Amount(abs(quantity))))
        return self

    def validate(self) -> None:
        """
        Performs validations on the instance.

        :raises AssertionError: If the journal entry is inconsistent.
        """
        ## Get total debit and credit amounts:
        total_debit = isum(i.amount for i in self.debits)
        total_credit = isum(i.amount for i in self.credits)

        ## Check:
        assert total_debit == total_credit, f"Total Debits and Credits are not equal: {total_debit} != {total_credit}"


class ReadJournalEntries(Protocol[_T]):
    """
    Type of functions which read journal entries from a source.
    """

    def __call__(self, period: DateRange) -> Iterable[JournalEntry[_T]]:
        pass



"""
:py:mod:`pypara` is a Python library for

- encoding currencies and monetary value objects,
- performing monetary arithmetic and conversions, and
- running rudimentary accounting operations.

Furthermore, there are some type convenience for general use.

The source code is organised in sub-packages and sub-modules.
"""

#: Defines the version of the package.
__version__ = "0.0.25"



from pypara.currencies import Currencies


def test_order() -> None:
    ## Get two currency instances:
    ccy1 = Currencies["USD"]
    ccy2 = Currencies["EUR"]

    ## Test:
    assert ccy1 == ccy1
    assert ccy1 > ccy2
    assert sorted([ccy1, ccy2]) == [ccy2, ccy1]



"""
This module provides definitions and functionality related to foreign-exchange conversions.
"""

__all__ = ["FXRate", "FXRateLookupError", "FXRateService"]

from abc import ABCMeta, abstractmethod
from decimal import Decimal
from typing import Iterable, NamedTuple, Optional, Tuple

from .commons.numbers import ONE, ZERO
from .commons.zeitgeist import Date
from .currencies import Currency


class FXRateLookupError(LookupError):
    """
    Provides an exception indicating that the foreign exchange rate is not found.
    """

    def __init__(self, ccy1: Currency, ccy2: Currency, asof: Date) -> None:
        """
        Initializes the foreign exchange rate lookup error.
        """
        ## Keep the slots:
        self.ccy1 = ccy1
        self.ccy2 = ccy2
        self.asof = asof

        ## Set the message:
        super().__init__(f"Foreign exchange rate for {ccy1}/{ccy2} not found as of {asof}")


class FXRate(NamedTuple):
    """
    Defines a foreign exchange (FX) rate model.

    Note that the constructor of this class is not safe: It does not check input. :method:`FXRate.of`, on the
    other hand provides a safer way of creating :class:`FXRate` instances.

    **Implementation Note:**

    I wanted to use an immutable, compact object model with fast creation and property access. Options were
    tweaked plain-vanilla Python class, NamedTuple and dataclasses.

    NamedTuple has slightly slower property access, whereby immutable dataclasses are slow for creation.

    Furthermore, as of the implementation of this class, mypy does not have proper dataclass support. Therefore,
    I am sticking to NamedTuple implementation.

    Last but not least, as objects are essentially tuples, indexed access to properties is possible and slightly
    faster.

    >>> import datetime
    >>> from decimal import Decimal
    >>> from pypara.currencies import Currencies
    >>> rate = FXRate(Currencies["EUR"], Currencies["USD"], datetime.date.today(), Decimal("2"))
    >>> ccy1, ccy2, date, value = rate
    >>> ccy1 == Currencies["EUR"]
    True
    >>> ccy2 == Currencies["USD"]
    True
    >>> date == datetime.date.today()
    True
    >>> value == Decimal("2")
    True
    """

    #: Defines the first currency of the FX rate.
    ccy1: Currency

    #: Defines the second currency of the FX rate.
    ccy2: Currency

    #: Defines the date the FX rate is effective as of.
    date: Date

    #: Defines the value of the FX rate.
    value: Decimal

    def __invert__(self) -> "FXRate":
        """
        Returns the inverted foreign exchange rate.

        >>> import datetime
        >>> from decimal import Decimal
        >>> from pypara.currencies import Currencies
        >>> nrate = FXRate(Currencies["EUR"], Currencies["USD"], datetime.date.today(), Decimal("2"))
        >>> rrate = FXRate(Currencies["USD"], Currencies["EUR"], datetime.date.today(), Decimal("0.5"))
        >>> ~nrate == rrate
        True
        """
        return FXRate(self[1], self[0], self[2], self[3] ** -1)

    @classmethod
    def of(cls, ccy1: Currency, ccy2: Currency, date: Date, value: Decimal) -> "FXRate":
        """
        Creates and returns an FX rate instance by validating arguments.

        >>> import datetime
        >>> from decimal import Decimal
        >>> from pypara.currencies import Currencies
        >>> urate = FXRate(Currencies["EUR"], Currencies["USD"], datetime.date.today(), Decimal("2"))
        >>> srate = FXRate.of(Currencies["EUR"], Currencies["USD"], datetime.date.today(), Decimal("2"))
        >>> urate == srate
        True
        """
        ## All argument must be of the respective specified type:
        if not isinstance(ccy1, Currency):
            raise ValueError("CCY/1 must be of type `Currency`.")
        if not isinstance(ccy2, Currency):
            raise ValueError("CCY/2 must be of type `Currency`.")
        if not isinstance(ccy1, Currency):
            raise ValueError("FX rate value must be of type `Decimal`.")
        if not isinstance(ccy1, Currency):
            raise ValueError("FX rate date must be of type `date`.")

        ## Check the value:
        if value <= ZERO:
            raise ValueError("FX rate value can not be equal to or less than `zero`.")

        ## Check consistency:
        if ccy1 == ccy2 and value != ONE:
            raise ValueError("FX rate to the same currency must be `one`.")

        ## Create and return the FX rate instance:
        return cls(ccy1, ccy2, date, value)


class FXRateService(metaclass=ABCMeta):
    """
    Provides an abstract class for serving foreign exchange rates.
    """

    #: Defines the default foreign exchange rate service for the runtime.
    default: Optional["FXRateService"] = None  # noqa: E704

    #: Defines an FX rate query tuple.
    TQuery = Tuple[Currency, Currency, Date]

    @abstractmethod
    def query(self, ccy1: Currency, ccy2: Currency, asof: Date, strict: bool = False) -> Optional[FXRate]:
        """
        Returns the foreign exchange rate of a given currency pair as of a given date.

        :param ccy1: The first currency of foreign exchange rate.
        :param ccy2: The second currency of foreign exchange rate.
        :param asof: Temporal dimension the foreign exchange rate is effective as of.
        :param strict: Indicates if we should raise a lookup error if that the foreign exchange rate can not be found.
        :return: The foreign exhange rate as a :class:`Decimal` instance or None.
        """
        pass

    @abstractmethod
    def queries(self, queries: Iterable[TQuery], strict: bool = False) -> Iterable[Optional[FXRate]]:
        """
        Returns foreign exchange rates for a given collection of currency pairs and dates.

        :param queries: An iterable of :class:`Currency`, :class:`Currency` and :class:`Temporal` tuples.
        :param strict: Indicates if we should raise a lookup error if that the foreign exchange rate can not be found.
        :return: An iterable of rates.
        """
        pass



import datetime
from decimal import Decimal

import pytest

from pypara.currencies import Currencies
from pypara.monetary import IncompatibleCurrencyError, Money, NonePrice, NoPrice, Price, SomePrice

## Define some currencies:
eur = Currencies["EUR"]
usd = Currencies["USD"]

## Defines some Decimal quantities:
zero = Decimal("0")
half = Decimal("0.5")
one = Decimal("1")
two = Decimal("2")

## Define some dates:
today = datetime.date.today()
yesterday = today - datetime.timedelta(days=1)


def test_implementation() -> None:
    ## Define instances:
    _price = Price()  # type: ignore
    sprice = SomePrice(usd, one, today)
    nprice = NonePrice()

    ## Check structure:
    assert _price.__slots__ == ()
    assert sprice.__slots__ == ()
    assert nprice.__slots__ == ()
    assert not hasattr(_price, "__dict__")
    assert not hasattr(sprice, "__dict__")
    assert not hasattr(nprice, "__dict__")

    ## Check types
    assert isinstance(Price.NA, Price)
    assert isinstance(Price.NA, NonePrice)
    assert not isinstance(Price.NA, SomePrice)
    assert not isinstance(Price.NA, Money)

    assert isinstance(_price, Price)
    assert not isinstance(_price, SomePrice)
    assert not isinstance(_price, NonePrice)

    assert isinstance(sprice, Price)
    assert isinstance(sprice, SomePrice)
    assert not isinstance(sprice, NonePrice)

    assert isinstance(nprice, Price)
    assert not isinstance(nprice, SomePrice)
    assert isinstance(nprice, NonePrice)


def test_of() -> None:
    assert Price.of(usd, one, None) == Price.NA
    assert Price.of(usd, None, today) == Price.NA
    assert Price.of(usd, one, None) == Price.NA
    assert Price.of(usd, one, today) == SomePrice(usd, one, today)


def test_is_equal() -> None:
    ## Vanilla:
    assert Price.NA.is_equal(NoPrice)
    assert Price.NA.is_equal(NonePrice())
    assert not Price.NA.is_equal(Price.of(usd, zero, today))
    assert Price.of(usd, zero, today).is_equal(Price.of(usd, zero, today))
    assert Price.of(usd, half, today).is_equal(Price.of(usd, half, today))
    assert not Price.of(usd, zero, today).is_equal(Price.of(eur, zero, today))
    assert not Price.of(usd, zero, today).is_equal(Price.of(usd, half, today))
    assert not Price.of(usd, zero, today).is_equal(Price.of(usd, zero, yesterday))

    ## With operator overload:
    assert Price.NA == NonePrice()
    assert Price.NA != Price.of(usd, zero, today)
    assert Price.of(usd, zero, today) == Price.of(usd, zero, today)
    assert Price.of(usd, half, today) == Price.of(usd, half, today)
    assert Price.of(usd, zero, today) != Price.of(eur, zero, today)
    assert Price.of(usd, zero, today) != Price.of(usd, half, today)
    assert Price.of(usd, zero, today) != Price.of(usd, zero, yesterday)


def test_to_boolean() -> None:
    ## Vanilla:
    assert not Price.NA.as_boolean()
    assert not Price.of(usd, zero, today).as_boolean()
    assert Price.of(usd, half, today).as_boolean()
    assert Price.of(usd, -half, today).as_boolean()

    ## With semantic overload
    assert not bool(Price.NA)
    assert not Price.of(usd, zero, today)
    assert Price.of(usd, half, today)
    assert Price.of(usd, -half, today)


def test_to_float() -> None:
    ## Vanilla:
    with pytest.raises(TypeError):
        Price.NA.as_float()
    assert Price.of(usd, half, today).as_float() == 0.5
    assert type(Price.of(usd, half, today).as_float()) == float

    ## With overload:
    with pytest.raises(TypeError):
        float(Price.NA)
    assert float(Price.of(usd, half, today)) == 0.5
    assert type(float(Price.of(usd, half, today))) == float


def test_to_integer() -> None:
    ## Vanilla:
    with pytest.raises(TypeError):
        int(Price.NA)
    assert int(Price.of(usd, half, today)) == 0
    assert type(int(Price.of(usd, half, today))) == int

    ## With overload:
    with pytest.raises(TypeError):
        Price.NA.as_integer()
    assert Price.of(usd, half, today).as_integer() == 0
    assert type(Price.of(usd, half, today).as_integer()) == int


def test_abs() -> None:
    ## Vanilla:
    assert Price.NA.abs() == Price.NA
    assert Price.of(usd, zero, today).abs() == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today).abs() == Price.of(usd, +one, today)
    assert Price.of(usd, +one, today).abs() == Price.of(usd, +one, today)

    ## With overload:
    assert abs(Price.NA) == Price.NA
    assert abs(Price.of(usd, zero, today)) == Price.of(usd, zero, today)
    assert abs(Price.of(usd, -one, today)) == Price.of(usd, +one, today)
    assert abs(Price.of(usd, +one, today)) == Price.of(usd, +one, today)


def test_negative() -> None:
    ## Vanilla:
    assert Price.NA.negative() == Price.NA
    assert Price.of(usd, zero, today).negative() == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today).negative() == Price.of(usd, +one, today)
    assert Price.of(usd, +one, today).negative() == Price.of(usd, -one, today)

    ## With overload:
    assert -Price.NA == Price.NA
    assert -Price.of(usd, zero, today) == Price.of(usd, zero, today)
    assert -Price.of(usd, -one, today) == Price.of(usd, +one, today)
    assert -Price.of(usd, +one, today) == Price.of(usd, -one, today)


def test_positive() -> None:
    ## Vanilla:
    assert Price.NA.positive() == Price.NA
    assert Price.of(usd, zero, today).positive() == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today).positive() == Price.of(usd, -one, today)
    assert Price.of(usd, +one, today).positive() == Price.of(usd, +one, today)

    ## With overload:
    assert +Price.NA == Price.NA
    assert +Price.of(usd, zero, today) == Price.of(usd, zero, today)
    assert +Price.of(usd, -one, today) == Price.of(usd, -one, today)
    assert +Price.of(usd, +one, today) == Price.of(usd, +one, today)


def test_round() -> None:
    ## Vanilla:
    assert Price.NA.round(2) == Price.NA
    assert Price.of(usd, zero, today).round(2) == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today).round(2) == Price.of(usd, -one, today)
    assert Price.of(usd, +one, today).round(2) == Price.of(usd, +one, today)

    ## Quick tests:
    assert Price.of(usd, Decimal("1.555"), today).round(2) == Price.of(usd, Decimal("1.56"), today)
    assert Price.of(usd, Decimal("1.545"), today).round(2) == Price.of(usd, Decimal("1.54"), today)

    ## With overload:
    assert round(Price.NA, 2) == Price.NA
    assert round(Price.of(usd, zero, today), 2) == Price.of(usd, zero, today)
    assert round(Price.of(usd, -one, today), 2) == Price.of(usd, -one, today)
    assert round(Price.of(usd, +one, today), 2) == Price.of(usd, +one, today)
    assert round(Price.of(usd, Decimal("1.555"), today), 2) == Price.of(usd, Decimal("1.56"), today)
    assert round(Price.of(usd, Decimal("1.545"), today), 2) == Price.of(usd, Decimal("1.54"), today)

    ## TODO: Following two are not really what round function signature says. mypy can't detect it!
    assert round(Price.of(usd, Decimal("1.4"), today)) == Price.of(usd, Decimal("1"), today)
    assert round(Price.of(usd, Decimal("1.5"), today)) == Price.of(usd, Decimal("2"), today)


def test_addition() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.add(Price.NA) == Price.NA
    assert Price.NA.add(Price.of(usd, zero, today)) == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today).add(Price.NA) == Price.of(usd, zero, today)

    ## Vanilla addition:
    assert Price.of(usd, zero, today).add(Price.of(usd, zero, today)) == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today).add(Price.of(usd, one, today)) == Price.of(usd, one, today)
    assert Price.of(usd, one, today).add(Price.of(usd, one, today)) == Price.of(usd, two, today)
    assert Price.of(usd, one, today).add(Price.of(usd, -one, today)) == Price.of(usd, zero, today)

    ## Carry dates forward:
    assert Price.of(usd, zero, today).add(Price.of(usd, one, yesterday)) == Price.of(usd, one, today)
    assert Price.of(usd, zero, yesterday).add(Price.of(usd, one, today)) == Price.of(usd, one, today)

    ## Incompatible currency errors:
    with pytest.raises(IncompatibleCurrencyError):
        Price.of(usd, zero, today).add(Price.of(eur, zero, today))

    ## Operator overload:
    assert Price.NA + Price.NA == Price.NA
    assert Price.NA + Price.of(usd, zero, today) == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today) + Price.NA == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today) + Price.of(usd, one, today) == Price.of(usd, one, today)


def test_scalar_addition() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.scalar_add(1) == Price.NA

    ## Vanilla addition:
    assert Price.of(usd, zero, today).scalar_add(1) == Price.of(usd, one, today)
    assert Price.of(usd, zero, today).scalar_add(1.0) == Price.of(usd, one, today)
    assert Price.of(usd, zero, today).scalar_add(one) == Price.of(usd, one, today)
    assert Price.of(usd, zero, today).scalar_add(-1) == Price.of(usd, -one, today)


def test_subtraction() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.subtract(Price.NA) == Price.NA
    assert Price.NA.subtract(Price.of(usd, zero, today)) == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today).subtract(Price.NA) == Price.of(usd, zero, today)

    ## Vanilla subtraction:
    assert Price.of(usd, zero, today).subtract(Price.of(usd, zero, today)) == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today).subtract(Price.of(usd, one, today)) == Price.of(usd, -one, today)
    assert Price.of(usd, one, today).subtract(Price.of(usd, one, today)) == Price.of(usd, zero, today)
    assert Price.of(usd, one, today).subtract(Price.of(usd, -one, today)) == Price.of(usd, two, today)

    ## Carry dates forward:
    assert Price.of(usd, zero, today).subtract(Price.of(usd, one, yesterday)) == Price.of(usd, -one, today)
    assert Price.of(usd, zero, yesterday).subtract(Price.of(usd, one, today)) == Price.of(usd, -one, today)

    ## Incompatible currency errors:
    with pytest.raises(IncompatibleCurrencyError):
        Price.of(usd, zero, today).subtract(Price.of(eur, zero, today))

    ## Operator overload:
    assert Price.of(usd, zero, today) - Price.of(usd, one, today) == Price.of(usd, -one, today)
    assert Price.NA - Price.NA == Price.NA
    assert Price.NA - Price.of(usd, zero, today) == Price.of(usd, zero, today)
    assert Price.of(usd, zero, today) - Price.NA == Price.of(usd, zero, today)


def test_scalar_subtraction() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.scalar_subtract(1) == Price.NA

    ## Vanilla subtraction:
    assert Price.of(usd, zero, today).scalar_subtract(1) == Price.of(usd, -one, today)
    assert Price.of(usd, zero, today).scalar_subtract(1.0) == Price.of(usd, -one, today)
    assert Price.of(usd, zero, today).scalar_subtract(one) == Price.of(usd, -one, today)
    assert Price.of(usd, zero, today).scalar_subtract(-1) == Price.of(usd, one, today)

    ## Operator overload:
    assert Price.of(usd, zero, today).scalar_subtract(1) == Price.of(usd, -one, today)
    assert Price.of(usd, zero, today).scalar_subtract(-1) == Price.of(usd, one, today)


def test_scalar_multiplication() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.multiply(1) == Price.NA

    ## Vanilla subtraction:
    assert Price.of(usd, one, today).multiply(1) == Price.of(usd, one, today)
    assert Price.of(usd, one, today).multiply(2) == Price.of(usd, two, today)
    assert Price.of(usd, -one, today).multiply(1) == Price.of(usd, -one, today)
    assert Price.of(usd, -one, today).multiply(2) == Price.of(usd, -two, today)

    ## Other types:
    assert Price.of(usd, one, today).multiply(1) == Price.of(usd, one, today)
    assert Price.of(usd, one, today).multiply(1.0) == Price.of(usd, one, today)
    assert Price.of(usd, one, today).multiply(one) == Price.of(usd, one, today)

    ## Operator overload:
    assert Price.NA * 1 == Price.NA
    assert Price.of(usd, one, today) * 1 == Price.of(usd, one, today)
    assert Price.of(usd, one, today) * 2 == Price.of(usd, two, today)
    assert Price.of(usd, -one, today) * 1 == Price.of(usd, -one, today)
    assert Price.of(usd, -one, today) * 2 == Price.of(usd, -two, today)


def test_monetary_multiplication() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.times(1) == Money.NA

    ## Vanilla multiplication:
    assert Price.of(usd, one, today).times(1) == Money.of(usd, one, today)
    assert Price.of(usd, one, today).times(2) == Money.of(usd, two, today)
    assert Price.of(usd, -one, today).times(1) == Money.of(usd, -one, today)
    assert Price.of(usd, -one, today).times(2) == Money.of(usd, -two, today)

    ## Other types:
    assert Price.of(usd, one, today).times(1) == Money.of(usd, one, today)
    assert Price.of(usd, one, today).times(1.0) == Money.of(usd, one, today)
    assert Price.of(usd, one, today).times(one) == Money.of(usd, one, today)


def test_division() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.divide(1) == Price.NA

    ## Vanilla subtraction:
    assert Price.of(usd, one, today).divide(1) == Price.of(usd, one, today)
    assert Price.of(usd, one, today).divide(2) == Price.of(usd, half, today)
    assert Price.of(usd, -one, today).divide(1) == Price.of(usd, -one, today)
    assert Price.of(usd, -one, today).divide(2) == Price.of(usd, -half, today)

    ## Various divisor types:
    assert Price.of(usd, one, today).divide(2) == Price.of(usd, half, today)
    assert Price.of(usd, one, today).divide(2.0) == Price.of(usd, half, today)
    assert Price.of(usd, one, today).divide(two) == Price.of(usd, half, today)

    ## Division by zero:
    assert Price.of(usd, one, today).divide(0) == Price.NA
    assert Price.of(usd, one, today).divide(zero) == Price.NA
    assert Price.of(usd, one, today).divide(0.0) == Price.NA

    ## Operator overload:
    assert Price.NA / 1 == Price.NA
    assert Price.of(usd, one, today) / 1 == Price.of(usd, one, today)
    assert Price.of(usd, one, today) / 2 == Price.of(usd, half, today)
    assert Price.of(usd, -one, today) / 1 == Price.of(usd, -one, today)
    assert Price.of(usd, -one, today) / 2 == Price.of(usd, -half, today)
    assert Price.of(usd, -one, today) / 0 == Price.NA


def test_floor_division() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.floor_divide(1) == Price.NA

    ## Vanilla subtraction:
    assert Price.of(usd, one, today).floor_divide(1) == Price.of(usd, one, today)
    assert Price.of(usd, one, today).floor_divide(2) == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today).floor_divide(1) == Price.of(usd, -one, today)
    assert Price.of(usd, -one, today).floor_divide(2) == Price.of(usd, zero, today)

    ## Various divisor types:
    assert Price.of(usd, one, today).floor_divide(2) == Price.of(usd, zero, today)
    assert Price.of(usd, one, today).floor_divide(2.0) == Price.of(usd, zero, today)
    assert Price.of(usd, one, today).floor_divide(two) == Price.of(usd, zero, today)

    ## Division by zero:
    assert Price.of(usd, one, today).floor_divide(0) == Price.NA
    assert Price.of(usd, one, today).floor_divide(zero) == Price.NA
    assert Price.of(usd, one, today).floor_divide(0.0) == Price.NA

    ## Operator overload:
    assert Price.NA / 1 == Price.NA
    assert Price.of(usd, one, today) // 1 == Price.of(usd, one, today)
    assert Price.of(usd, one, today) // 2 == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today) // 1 == Price.of(usd, -one, today)
    assert Price.of(usd, -one, today) // 2 == Price.of(usd, zero, today)
    assert Price.of(usd, -one, today) // 0 == Price.NA


def test_comparisons() -> None:
    ## First use `Price.NA`s:
    assert not (Price.NA < Price.NA)
    assert Price.NA <= Price.NA
    assert not (Price.NA > Price.NA)
    assert Price.NA >= Price.NA

    ## Try mixed:
    assert Price.NA < Price.of(usd, -one, today)
    assert Price.NA <= Price.of(usd, -one, today)
    assert not (Price.NA > Price.of(usd, -one, today))
    assert not (Price.NA >= Price.of(usd, -one, today))

    ## ... and:
    assert not (Price.of(usd, -one, today) < Price.NA)
    assert not (Price.of(usd, -one, today) <= Price.NA)
    assert Price.of(usd, -one, today) > Price.NA
    assert Price.of(usd, -one, today) >= Price.NA

    ## With defined values:
    assert not (Price.of(usd, zero, today) < Price.of(usd, zero, today))
    assert Price.of(usd, zero, today) <= Price.of(usd, zero, today)
    assert not (Price.of(usd, zero, today) > Price.of(usd, zero, today))
    assert Price.of(usd, zero, today) >= Price.of(usd, zero, today)

    ## ... and:
    assert Price.of(usd, zero, today) < Price.of(usd, one, today)
    assert Price.of(usd, zero, today) <= Price.of(usd, one, today)
    assert not (Price.of(usd, zero, today) > Price.of(usd, one, today))
    assert not (Price.of(usd, zero, today) >= Price.of(usd, one, today))

    ## ... and:
    assert not (Price.of(usd, one, today) < Price.of(usd, zero, today))
    assert not (Price.of(usd, one, today) <= Price.of(usd, zero, today))
    assert Price.of(usd, one, today) > Price.of(usd, zero, today)
    assert Price.of(usd, one, today) >= Price.of(usd, zero, today)


def test_with() -> None:
    ## First use `Price.NA`s:
    assert Price.NA.with_ccy(usd) == Price.NA
    assert Price.NA.with_qty(one) == Price.NA
    assert Price.NA.with_dov(today) == Price.NA

    ## Now with some:
    assert Price.of(usd, zero, today).with_ccy(eur) == Price.of(eur, zero, today)
    assert Price.of(usd, zero, today).with_qty(one) == Price.of(usd, one, today)
    assert Price.of(usd, zero, today).with_dov(yesterday) == Price.of(usd, zero, yesterday)


def test_types() -> None:
    assert isinstance(Price.NA, Price)
    assert isinstance(Price.NA, NonePrice)
    assert not isinstance(Price.NA, SomePrice)
    assert not isinstance(Price.NA, Money)



__all__ = ["identity"]

from typing import Iterable, List, TypeVar

#: A named, generic type variable.
_T = TypeVar("_T")


def identity(x: _T) -> _T:
    """
    Provides the identity function.

    :param x: Any value of the generic type.
    :return: The value consumed.

    >>> identity(1)
    1
    """
    return x


def chunk(lst: List[_T], n: int) -> Iterable[List[_T]]:
    """
    Splits given list in chunks of given size and returns them in an iterable.

    :param lst: List of items.
    :param n: Chunks size.
    :return: Iterable of item lists.

    >>> list(chunk([], 1))
    []
    >>> list(chunk([], 2))
    []
    >>> list(chunk([1, 2, 3, 4], 1))
    [[1], [2], [3], [4]]
    >>> list(chunk([1, 2, 3, 4], 2))
    [[1, 2], [3, 4]]
    >>> list(chunk([1, 2, 3, 4, 5], 2))
    [[1, 2], [3, 4], [5]]
    """
    for i in range(0, len(lst), n):
        yield lst[i : i + n]  # noqa: E203



"""
This module provides data definitions and related functionality for preparing general ledgers.
"""

__all__ = [
    "GeneralLedger",
    "GeneralLedgerProgram",
    "InitialBalances",
    "Ledger",
    "LedgerEntry",
    "ReadInitialBalances",
    "build_general_ledger",
    "compile_general_ledger_program",
]

import datetime
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Dict, Generic, Iterable, List, Optional, Protocol, TypeVar

from ..commons.numbers import Amount, Quantity
from ..commons.zeitgeist import DateRange
from .accounts import Account
from .generic import Balance
from .journaling import JournalEntry, Posting, ReadJournalEntries

#: Defines a generic type variable.
_T = TypeVar("_T")


#: Initial balances:
InitialBalances = Dict[Account, Balance]


@dataclass
class LedgerEntry(Generic[_T]):
    """
    Provides a ledger entry model.
    """

    #: Ledger the entry belongs to.
    ledger: "Ledger[_T]"

    #: Posting of the ledger entry.
    posting: Posting[_T]

    #: Balance of the ledger entry.
    balance: Quantity

    @property
    def date(self) -> datetime.date:
        """
        Date of the ledger entry.
        """
        return self.posting.date

    @property
    def description(self) -> str:
        """
        Description of the ledger entry.
        """
        return self.posting.journal.description

    @property
    def amount(self) -> Amount:
        """
        Amount of the ledger entry.
        """
        return self.posting.amount

    @property
    def cntraccts(self) -> List[Account]:
        """
        Counter accounts for the ledger entry.
        """
        return [p.account for p in self.posting.journal.postings if p.direction != self.posting.direction]

    @property
    def is_debit(self) -> bool:
        """
        Indicates if the ledger entry is a debit.
        """
        return self.posting.is_debit

    @property
    def is_credit(self) -> bool:
        """
        Indicates if the ledger entry is a credit.
        """
        return self.posting.is_credit

    @property
    def debit(self) -> Optional[Amount]:
        """
        Returns the debit amount, if any.
        """
        return self.amount if self.is_debit else None

    @property
    def credit(self) -> Optional[Amount]:
        """
        Returns the credit amount, if any.
        """
        return self.amount if self.is_credit else None


@dataclass
class Ledger(Generic[_T]):
    """
    Provides an account ledger model.
    """

    #: Account of the ledger.
    account: Account

    #: Initial balance of the ledger.
    initial: Balance

    #: Ledger entries.
    entries: List[LedgerEntry[_T]] = field(default_factory=list, init=False)

    @property
    def _last_balance(self) -> Quantity:
        """
        Returns the last balance.
        """
        try:
            return self.entries[-1].balance
        except IndexError:
            return self.initial.value

    def add(self, posting: Posting[_T]) -> LedgerEntry[_T]:
        """
        Adds a new ledger entry.

        :param posting: Posting the ledger entry is based on.
        :return: The new ledger entry.
        """
        ## Create the ledger entry.
        entry = LedgerEntry(self, posting, Quantity(self._last_balance + posting.amount * posting.direction.value))

        ## Add to the buffer:
        self.entries.append(entry)

        ## Done, return:
        return entry


@dataclass
class GeneralLedger(Generic[_T]):
    """
    Provides a general ledger model.
    """

    #: Accounting period.
    period: DateRange

    #: Individual account ledgers of the general ledger.
    ledgers: Dict[Account, Ledger[_T]]


def build_general_ledger(
    period: DateRange, journal: Iterable[JournalEntry[_T]], initial: InitialBalances
) -> GeneralLedger[_T]:
    """
    Builds a general ledger.

    :param period: Accounting period.
    :param journal: All available journal entries.
    :param initial: Opening balances for terminal accounts, if any.
    :return: A :py:class:`GeneralLedger` instance.
    """
    ## Initialize ledgers buffer as per available initial balances:
    ledgers: Dict[Account, Ledger[_T]] = {a: Ledger(a, b) for a, b in initial.items()}

    ## Iterate over journal postings and populate ledgers:
    for posting in (p for j in journal for p in j.postings if period.since <= j.date <= period.until):
        ## Check if we have the ledger yet, and create if not:
        if posting.account not in ledgers:
            ledgers[posting.account] = Ledger(posting.account, Balance(period.since, Quantity(Decimal(0))))

        ## Add the posting to the ledger:
        ledgers[posting.account].add(posting)

    ## Done, return general ledger.
    return GeneralLedger(period, ledgers)


class ReadInitialBalances(Protocol):
    """
    Type of functions which reads and returns initial balances.
    """

    def __call__(self, period: DateRange) -> InitialBalances:
        pass


class GeneralLedgerProgram(Protocol[_T]):
    """
    Type definition of the program which builds general ledger.
    """

    def __call__(self, period: DateRange) -> GeneralLedger[_T]:
        pass


def compile_general_ledger_program(
    read_initial_balances: ReadInitialBalances,
    read_journal_entries: ReadJournalEntries[_T],
) -> GeneralLedgerProgram[_T]:
    """
    Consumes implementations of the algebra and returns a program which consumes opening and closing dates and produces
    a general ledger.

    :param read_initial_balances: Algebra implementation which reads initial balances.
    :param read_journal_entries: Algebra implementation which reads journal entries.
    :return: A function which consumes opening and closing dates and produces a general ledger
    """

    def _program(period: DateRange) -> GeneralLedger[_T]:
        """
        Consumes the opening and closing dates and produces a general ledger.

        :param period: Accounting period.
        :return: A general ledger.
        """
        ## Get initial balances as of the end of previous financial period:
        initial_balances = read_initial_balances(period)

        ## Read journal entries and post each of them:
        journal_entries = read_journal_entries(period)

        ## Build the general ledger and return:
        return build_general_ledger(period, journal_entries, initial_balances)

    ## Return the compiled program.
    return _program



"""
This module defines type definitions and functions for various purposes.
"""

__all__ = ["Guid", "makeguid"]

from typing import NewType
from uuid import uuid4

#: Defines a new-type for globally-unique identifiers.
Guid = NewType("Guid", str)


def makeguid() -> Guid:
    """
    Creates a new :py:class:`Guid`.

    :return: :py:class:`Guid` instance.

    >>> isinstance(makeguid(), str)  ## During static analysis, it is `Guid`.
    True
    """
    return Guid(uuid4().hex)
